from surreal.session import (
    Config,
    LOCAL_SESSION_CONFIG,
    BASE_LEARNER_CONFIG,
    BASE_ENV_CONFIG
    )
from surreal.agent import PPOAgent
from surreal.learner import PPOLearner
from surreal.replay import FIFOReplay
from surreal.launch import SurrealDefaultLauncher
from surreal.env import make_env
import argparse


PPO_DEFAULT_LEARNER_CONFIG = Config({
    'model': {
        'convs': [],  # this can wait until TorchX
        'actor_fc_hidden_sizes': [300, 200],
        'critic_fc_hidden_sizes': [300, 200],
        'cnn_feature_dim': 256,
        'use_layernorm': False,
    },
    'algo': {
        # base configs
        # 'agent_class': 'PPOAgent',
        # 'learner_class': 'PPOLearner',
        # 'experience': 'ExpSenderWrapperMultiStepMovingWindowWithInfo',
        'use_z_filter': False,
        'use_r_filter': False,
        'gamma': .99,
        'n_step': 25,  # 10 for without RNN
        'stride': 20,  # 10 for without RNN
        'network': {
            'lr_actor': 1e-4,
            'lr_critic': 1e-4,
            'clip_actor_gradient': True,
            'actor_gradient_norm_clip': 10.,
            'clip_critic_gradient': True,
            'critic_gradient_norm_clip': 10.,
            'actor_regularization': 0.0,
            'critic_regularization': 0.0,
            'anneal': {
                'lr_scheduler': "LinearWithMinLR",
                'frames_to_anneal': 5e6,
                'lr_update_frequency': 100,
                'min_lr': 1e-4,
            },
        },

        # ppo specific parameters:
        'ppo_mode': 'adapt',
        'advantage':{
            'norm_adv': True,
            'lam': 1.0,
            'reward_scale': 1.0,
        },
        'rnn': {
            'if_rnn_policy': True,
            'rnn_hidden': 100,
            'rnn_layer': 1,
            'horizon': 5,
        },
        'consts': {
            'init_log_sig': -1.0,
            'log_sig_range': 0,
            'epoch_policy': 10,
            'epoch_baseline': 10,
            'adjust_threshold': (0.5, 2.0),  # threshold to magnify clip epsilon
            'kl_target': 0.02,  # target KL divergence between before and after
        },
        'adapt_consts': {
            'kl_cutoff_coeff': 500,  # penalty coeff when kl large
            'beta_init': 1.0,  # original beta
            'beta_range': (1/35.0, 35.0),  # range of the adapted penalty factor
            'scale_constant': 1.5,
        },
        'clip_consts': {
            'clip_epsilon_init': 0.2,  # factor of clipped loss
            'clip_range': (0.05, 0.3),  # range of the adapted penalty factor
            'scale_constant': 1.2,
        },

    },
    'replay': {
        # 'replay_class': 'FIFOReplay',
        'batch_size': 64,
        'memory_size': 96,
        'sampling_start_size': 64,
        'replay_shards': 1,
    },
    'parameter_publish': {
        'exp_interval': 4096,
    },
})
PPO_DEFAULT_LEARNER_CONFIG.extend(BASE_LEARNER_CONFIG)

PPO_DEFAULT_ENV_CONFIG = Config({
    'env_name': '',
    'action_repeat': 10,
    'pixel_input': False,
    'use_grayscale': False,
    'use_depth': False,
    'frame_stacks': 1,
    'sleep_time': 0,
    'video': {
        'record_video': False,
        'save_folder': None,
        'max_videos': 500,
        'record_every': 5,
    },
    'observation': {
        'pixel': ['camera0'],
        'low_dim':['position', 'velocity', 'proprio', 'cube_pos', 'cube_quat', 'gripper_to_cube', 'low-dim'],
    },
    'eval_mode': {
        'demonstration': None
    },
    'demonstration': {
        'use_demo': False,
        'adaptive': True,
        # params for open loop reverse curriculum
        'increment_frequency': 100,
        'sample_window_width': 25,
        'increment': 25,

        # params for adaptive curriculum
        'mixing': ['random'],
        'mixing_ratio': [1.0],
        'ratio_step': [0.0],
        'improve_threshold': 0.1,
        'curriculum_length': 50,
        'history_length': 20,
    },
    'limit_episode_length': 500,
    'stochastic_eval': True,
})
PPO_DEFAULT_ENV_CONFIG.extend(BASE_ENV_CONFIG)

PPO_DEFAULT_SESSION_CONFIG = Config({
    'folder': '_str_',
    'tensorplex': {
        'update_schedule': {
            # for TensorplexWrapper:
            'training_env': 20,  # env record every N episodes
            'eval_env': 5,
            'eval_env_sleep': 2,  # throttle eval by sleep n seconds
            # for manual updates:
            'agent': 50,  # agent.update_tensorplex()
            'learner': 20,  # learner.update_tensorplex()
        },
    },
    'agent': {
        'fetch_parameter_mode': 'step',
        'fetch_parameter_interval': 250,  # 10 for without RNN
        'num_gpus': 0,
    },
    'sender': {
        'flush_iteration': 3,
    },
    'learner': {
        'num_gpus': 0,
    },
    'replay': {
        'max_puller_queue': 3,
        'max_prefetch_queue': 1,
    },
    'checkpoint': {
        'learner': {
            'mode': 'history',
            'periodic': 1000, # Save every 1000 steps
            'min_interval': 15 * 60, # No checkpoint less than 15 min apart.
        },
    },
})
PPO_DEFAULT_SESSION_CONFIG.extend(LOCAL_SESSION_CONFIG)


class PPOLauncher(SurrealDefaultLauncher):
    def __init__(self):
        learner_class = PPOLearner
        agent_class = PPOAgent
        replay_class = FIFOReplay
        learner_config = PPO_DEFAULT_LEARNER_CONFIG
        env_config = PPO_DEFAULT_ENV_CONFIG
        session_config = PPO_DEFAULT_SESSION_CONFIG
        super().__init__(agent_class,
                         learner_class,
                         replay_class,
                         session_config,
                         env_config,
                         learner_config)

    def setup(self, argv):
        parser = argparse.ArgumentParser()
        parser.add_argument('--env', type=str, required=True, help='name of the environment')
        parser.add_argument('--num-agents', type=int, required=True, help='number of agents used')
        parser.add_argument('--num-gpus', type=int, default=0,
                            help='number of GPUs to use, 0 for CPU only.')
        parser.add_argument('--agent-num-gpus', type=int, default=0,
                            help='number of GPUs to use for agent, 0 for CPU only.')
        parser.add_argument('--restore-folder', type=str, default=None,
                            help='folder containing checkpoint to restore from')
        parser.add_argument('--experiment-folder', required=True,
                            help='session_config.folder that has experiment'
                            'files like checkpoint and logs')
        parser.add_argument('--agent-batch', type=int, default=1,
                            help='how many agents/evals per batch')
        parser.add_argument('--unit-test', action='store_true',
                            help='Set config values to settings that can run locally for unit testing')

        args = parser.parse_args(args=argv)

        self.env_config.env_name = args.env
        _, self.env_config = make_env(self.env_config)

        self.session_config.folder = args.experiment_folder
        self.session_config.agent.num_gpus = args.agent_num_gpus
        self.session_config.learner.num_gpus = args.num_gpus
        if args.restore_folder is not None:
            self.session_config.checkpoint.restore = True
            self.session_config.checkpoint.restore_folder = args.restore_folder
        self.agent_batch_size = args.agent_batch
        self.eval_batch_size = args.agent_batch

        if args.unit_test:
            self.learner_config.replay.batch_size = 2
            self.learner_config.replay.sampling_start_size = 2


def main():
    PPOLauncher().main()


if __name__ == '__main__':
    main()


'''
    Specific hyperparameters For Cheetah v. Hopper:
        * ADAPT
        * gamma: 0.995
        * lam: 0.97
        * kltarget: .01
        * update count: 10
        * kl_cutoff_coeff: 50
        * fetch_parameter_interval: 100
        * release_interval: 4096
        * zfilter: True/False
        * n_step: 30/15
        * stride: 20/10
        * initial learning rate: 3e-4/1e-4
        * annealed final learning rate: 5e-5
        * actor_gradient_norm_clip: 1./1.
        * critic_gradient_norm_clip: 5./1.
        * reward_scale: 1.0/0.01
        * init_log_sig: -1 / -2
        * log_sig_range: 0 / 0.5
'''



"""
Defines the LaunchSettings class that holds all the
information one needs to launch a component of surreal
"""
import time
import os
import sys
import subprocess
from argparse import ArgumentParser
import numpy as np
from tensorplex import Loggerplex
from tensorplex import Tensorplex
from surreal.distributed import ShardedParameterServer
from surreal.replay import ShardedReplay
import surreal.utils as U
import faulthandler
faulthandler.enable()


class Launcher:
    """
        Launchers are shared entrypoint for surreal experiments.
        Launchers define a main function that takes commandline
        arguments in the following way.
        `python launch_ppo.py <component_name> -- [additional_args]`
        component_name defines which part of the experiment should be
        run in this process
        [additional_args] should be shared among all involved processes
        to define behavior globally
    """
    def main(self):
        """
        The main function to be called
        ```
        if __name__ == '__main__':
            launcher = Launcher()
            launcher.main()
        ```
        """
        argv = sys.argv[1:]
        parser_args = argv
        config_args = []
        if '--' in argv:
            index = argv.index('--')
            parser_args = argv[:index]
            config_args = argv[index + 1:]
        parser = ArgumentParser(description='launch a surreal component')
        parser.add_argument('component_name',
                            type=str,
                            help='which component to launch')
        args = parser.parse_args(parser_args)

        self.config_args = config_args

        self.setup(config_args)
        self.launch(args.component_name)

    def launch(self, component_name):
        """Launches the specific component

        Args:
            component_name(str): the process to launch
        """
        raise NotImplementedError

    def setup(self, args):
        """
        Sets up the related states for running components

        Args:
            args: A list of commandline arguments provided.
        """
        pass


class SurrealDefaultLauncher(Launcher):
    """
        The default launcher instance of surreal.
    """
    def __init__(self,
                 agent_class,
                 learner_class,
                 replay_class,
                 session_config,
                 env_config,
                 learner_config,
                 eval_mode='eval_stochastic',
                 agent_batch_size=8,
                 eval_batch_size=8,
                 render=False):
        """
        Setup an surreal experiment

        Args:
            agent_class: The Agent subclass to run for agents
            learner_class: The Agent subclass to run for evals
            replay_class: The Replay subclass to run for replays
            session_config: configs passed to all components
            env_config: configs passed to all components
            learner_config: configs passed to all components
            eval_mode: whether evals should be deterministic or
                stochastic. 'eval_deterministic'/'eval_stochastic'
                (default: {'eval_stochastic'})
            agent_batch_size: When running batch_agent,
                how many agents to fork. (default: {8})
            eval_batch_size: When running batch_agent,
                how many evals to fork. (default: {8})
            render: Whether evals should render (default: {False})
        """
        self.agent_class = agent_class
        self.learner_class = learner_class
        self.replay_class = replay_class
        self.session_config = session_config
        self.env_config = env_config
        self.learner_config = learner_config

        self.eval_mode = eval_mode
        self.render = render
        self.agent_batch_size = agent_batch_size
        self.eval_batch_size = eval_batch_size

    def launch(self, component_name_in):
        """
            Launches a surreal experiment

        Args:
            component_name: Allowed components:
                                agent-{*},
                                agents-{*},
                                eval-{*},
                                evals-{*},
                                replay,
                                learner,
                                ps,
                                tensorboard,
        """
        if '-' in component_name_in:
            component_name, component_id = component_name_in.split('-')
            component_id = int(component_id)
        else:
            component_name = component_name_in
            component_id = None

        if component_name == 'agent':
            self.run_agent(agent_id=component_id)
        elif component_name == 'agents':
            agent_ids = self.get_agent_batch(component_id)
            self.run_agent_batch(agent_ids)
        elif component_name == 'eval':
            self.run_eval(eval_id=component_id,
                          mode=self.eval_mode,
                          render=self.render)
        elif component_name == 'evals':
            eval_ids = self.get_eval_batch(component_id)
            self.run_eval_batch(eval_ids, self.eval_mode, self.render)
        elif component_name == 'learner':
            self.run_learner()
        elif component_name == 'ps':
            self.run_ps()
        elif component_name == 'replay':
            self.run_replay()
        # elif component_name == 'replay_loadbalancer':
        #     self.run_replay_loadbalancer()
        # elif component_name == 'replay_worker':
        #     self.run_replay_worker(replay_id=component_id)
        elif component_name == 'tensorboard':
            self.run_tensorboard()
        elif component_name == 'tensorplex':
            self.run_tensorplex()
        elif component_name == 'loggerplex':
            self.run_loggerplex()
        else:
            raise ValueError('Unexpected component {}'.format(component_name))

    def run_component(self, component_name):
        return subprocess.Popen([sys.executable, '-u',
                                 sys.argv[0],
                                 component_name,
                                 '--'] + self.config_args)

    def run_agent(self, agent_id):
        """
            Launches an agent process with agent_id

        Args:
            agent_id (int): agent's id
            iterations (int): if not none, the number of episodes to run before exiting
        """
        agent = self.setup_agent(agent_id)
        agent.main_agent()

    def setup_agent(self, agent_id):
        """'

        Same as launch_agent, but instead of running agent.main()
        infinite loop, returns the agent instance

        Args:
            agent_id: [description]
        """
        np.random.seed(int(time.time() * 100000 % 100000))

        session_config, learner_config, env_config = \
            self.session_config, self.learner_config, self.env_config
        agent_mode = 'training'

        # env, env_config = make_env(env_config)

        agent = self.agent_class(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            agent_id=agent_id,
            agent_mode=agent_mode,
        )

        return agent

    def run_agent_batch(self, agent_ids):
        """
            Launches multiple agent processes with agent_id in agent_ids
            Useful when you want agents to share a GPU

        Args:
            agent_ids (list(int)): each agent's id
        """
        agents = []
        for agent_id in agent_ids:
            component_name = 'agent-{}'.format(agent_id)
            agent = self.run_component(component_name)
            agents.append(agent)
        U.wait_for_popen(agents)

    def get_agent_batch(self, batch_id):
        """
            Returns the agent_ids corresponding to batch_id

        Args:
            batch_id: index of batch

        Returns:
            agent_ids (list): ids of the agents in the batch
        """
        return range(self.agent_batch_size * int(batch_id),
                     self.agent_batch_size * (int(batch_id) + 1))

    def run_eval(self, eval_id, mode, render):
        """
            Launches an eval processes with id eval_id

        Args:
            eval_id (int): eval agent's id
            mode: eval_deterministic or eval_stochastic
            render: see run_eval
        """
        np.random.seed(int(time.time() * 100000 % 100000))

        session_config, learner_config, env_config = \
            self.session_config, self.learner_config, self.env_config

        # env, env_config = make_env(env_config, 'eval')

        agent_mode = mode
        assert agent_mode != 'training'

        agent_class = self.agent_class
        agent = agent_class(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            agent_id=eval_id,
            agent_mode=agent_mode,
            render=render
        )

        agent.main_eval()

    def run_eval_batch(self, eval_ids, mode, render):
        """
            Launches multiple eval processes with agent_id in agent_ids
            Useful when you want agents to share a GPU

        Args:
            eval_ids (list(int)): each eval agent's id
            mode:
            render: see run_eval
        """
        evals = []
        for eval_id in eval_ids:
            component_name = 'eval-{}'.format(eval_id)
            agent = self.run_component(component_name)
            evals.append(agent)
        U.wait_for_popen(evals)

    def get_eval_batch(self, batch_id):
        """
            Returns the eval_ids corresponding to batch_id

        Args:
            batch_id: index of batch

        Returns:
            eval_ids (list): ids of the agents in the batch
        """
        return range(self.eval_batch_size * int(batch_id),
                     self.eval_batch_size * int(batch_id) + 1)

    def run_learner(self, iterations=None):
        """
            Launches the learner process.
            Learner consumes experience from replay
            and publishes experience to parameter server
        """
        learner = self.setup_learner()
        learner.main()

    def setup_learner(self):
        """
            Same as run_learner, but returns the Learner instance
            instead of calling learner.main_loop()
        """
        session_config, learner_config, env_config = \
            self.session_config, self.learner_config, self.env_config

        learner_class = self.learner_class
        learner = learner_class(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config
        )

        return learner

    def run_ps(self):
        """
            Lauches the parameter server process.
            Serves parameters to agents
        """
        ps_config = self.session_config.ps

        server = ShardedParameterServer(shards=ps_config.shards)

        server.launch()
        server.join()

    def run_replay(self):
        """
            Launches the replay process.
            Replay collects experience from agents
            and serve them to learner
        """
        session_config, learner_config, env_config = \
            self.session_config, self.learner_config, self.env_config

        sharded = ShardedReplay(learner_config=learner_config,
                                env_config=env_config,
                                session_config=session_config,
                                replay_class=self.replay_class)

        sharded.launch()
        sharded.join()

    def run_tensorboard(self):
        """
            Launches a tensorboard process
        """
        folder = os.path.join(self.session_config.folder, 'tensorboard')
        tensorplex_config = self.session_config.tensorplex
        cmd = ['tensorboard',
               '--logdir', folder,
               '--port', str(tensorplex_config.tensorboard_port)]
        subprocess.call(cmd)

    def run_tensorplex(self):
        """
            Launches a tensorplex process.
            It receives data from multiple sources and
            send them to tensorboard.
        """
        folder = os.path.join(self.session_config.folder, 'tensorboard')
        tensorplex_config = self.session_config.tensorplex

        tensorplex = Tensorplex(
            folder,
            max_processes=tensorplex_config.max_processes,
        )

        """
            Tensorboard categories:
                learner/replay/eval: algorithmic level, e.g. reward, ...
                ***-core: low level metrics, i/o speed, computation time, etc.
                ***-system: Metrics derived from raw metric data in core,
                    i.e. exp_in/exp_out
        """
        (tensorplex
         .register_normal_group('learner')
         .register_indexed_group('agent', tensorplex_config.agent_bin_size)
         .register_indexed_group('eval', 4)
         .register_indexed_group('replay', 10))

        port = os.environ['SYMPH_TENSORPLEX_PORT']
        tensorplex.start_server(port=port)

    def run_loggerplex(self):
        """
            Launches a loggerplex server.
            It helps distributed logging.
        """
        folder = self.session_config.folder
        loggerplex_config = self.session_config.loggerplex

        loggerplex = Loggerplex(
            os.path.join(folder, 'logs'),
            level=loggerplex_config.level,
            overwrite=loggerplex_config.overwrite,
            show_level=loggerplex_config.show_level,
            time_format=loggerplex_config.time_format
        )
        port = os.environ['SYMPH_LOGGERPLEX_PORT']
        loggerplex.start_server(port)



"""
    Defines the parameter publishing mechanism that propagates
        updated parameters from the learner to agents
"""
import time
from multiprocessing import Process
import os
from caraml.zmq import (
    ZmqProxyThread,
    ZmqPub,
    ZmqSub,
    ZmqServer,
    ZmqClient,
    ZmqTimeoutError)
import surreal.utils as U
from surreal.distributed.module_dict import ModuleDict
# TODO: better logging for this file


class ParameterPublisher(object):
    """
        Publishes parameters from the learner side
        Using ZmqPub socket
    """
    def __init__(self, port, module_dict):
        """
        Args:
            port: the port connected to the pub socket
            module_dict: ModuleDict object that exposes model parameters
        """
        self._publisher = ZmqPub(
            host='*',
            port=port,
            serializer=U.serialize,
        )
        if not isinstance(module_dict, ModuleDict):
            module_dict = ModuleDict(module_dict)
        self._module_dict = module_dict

    def publish(self, iteration, message=''):
        """
        Called by learner. Publishes model parameters with additional info

        Args:
            iteration: current learning iteration
            message: any U.serialize serializable data
        """
        binary = self._module_dict.dumps()
        info = {
            'time': time.time(),
            'iteration': iteration,
            'message': message,
            'hash': U.binary_hash(binary)
        }
        self._publisher.pub(topic='ps', data=(binary, info))


class ShardedParameterServer(object):
    """
        Runs multiple parameter servers in parallel
    """
    def __init__(self, shards):
        self.shards = shards

        # Serving parameter to agents
        self.frontend_port = os.environ['SYMPH_PS_FRONTEND_PORT']
        self.backend_port = os.environ['SYMPH_PS_BACKEND_PORT']
        self.serving_frontend_add = "tcp://*:{}".format(self.frontend_port)
        self.serving_backend_add = "tcp://*:{}".format(self.backend_port)

        # Subscribing to learner published parameters
        self.publisher_host = os.environ['SYMPH_PARAMETER_PUBLISH_HOST']
        self.publisher_port = os.environ['SYMPH_PARAMETER_PUBLISH_PORT']

        self.proxy = None
        self.workers = []

    def launch(self):
        """
            Runs load balancing proxy thread
                and self.shards ParameterServer processes
            Returns after all threads and processes are running
        """
        self.proxy = ZmqProxyThread(in_add=self.serving_frontend_add,
                                    out_add=self.serving_backend_add,
                                    pattern='router-dealer')
        self.proxy.start()

        self.workers = []
        for i in range(self.shards):
            worker = ParameterServer(
                publisher_host=self.publisher_host,
                publisher_port=self.publisher_port,
                serving_host='localhost',
                serving_port=self.backend_port,
                load_balanced=True,
            )
            worker.start()
            self.workers.append(worker)

    def join(self):
        """
            Wait for all parameter server workers to exit
                (Currently this means they crashed)
            Note that proxy is a daemon thread and doesn't need waiting
        """
        for i, worker in enumerate(self.workers):
            worker.join()
            U.report_exitcode(worker.exitcode, 'ps-{}'.format(i))


class ParameterServer(Process):
    """
        Standalone script for PS node that runs in an infinite loop.
        The ParameterServer subscribes to learner to get the latest
            model parameters and serves these parameters to agents
        It implements a simple hash based caching mechanism to avoid
            serving duplicate parameters to agent
    """
    def __init__(self,
                 publisher_host,
                 publisher_port,
                 serving_host,
                 serving_port,
                 load_balanced=False,):
        """
        Args:
            publisher_host, publisher_port: where learner publish parameters
            serving_host, serving_port: where to serve parameters to agents
            load_balanced: whether multiple parameter servers are sharing the
                same address
        """
        Process.__init__(self)
        self.publisher_host = publisher_host
        self.publisher_port = publisher_port
        self.serving_host = serving_host
        self.serving_port = serving_port
        self.load_balanced = load_balanced
        # storage
        self.parameters = None
        self.param_info = None
        # threads
        self._subscriber = None
        self._server = None
        self._subscriber_thread = None
        self._server_thread = None

    def run(self):
        """
            Run relative threads and wait until they finish (due to error)
        """
        self._subscriber = ZmqSub(
            host=self.publisher_host,
            port=self.publisher_port,
            # handler=self._set_storage,
            topic='ps',
            deserializer=U.deserialize,
        )
        self._server = ZmqServer(
            host=self.serving_host,
            port=self.serving_port,
            # handler=self._handle_agent_request,
            serializer=U.serialize,
            deserializer=U.deserialize,
            bind=not self.load_balanced,
        )
        self._subscriber_thread = self._subscriber.start_loop(
            handler=self._set_storage,
            blocking=False)
        self._server_thread = self._server.start_loop(
            handler=self._handle_agent_request,
            blocking=False)
        print('Parameter server started')

        self._subscriber_thread.join()
        self._server_thread.join()

    def _set_storage(self, data):
        self.parameters, self.param_info = data

    def _handle_agent_request(self, request):
        """
            Reply to agents' request for parameters

        Args:
            request: 3 types
             - "info": (None, info)
             - "parameter": (param, info)
             - "parameter:<agent-hash>":
                returns (None, None) if no parameter has been published
                returns (None, info) if the hash
                    of server side parameters is the same as the agent's
                otherwise returns (param, info)
        """
        if request == 'info':
            return None, self.param_info
        elif request.startswith('parameter'):
            if self.parameters is None:
                return None, None
            if ':' in request:
                _, last_hash = request.split(':', 1)
                current_hash = self.param_info['hash']
                if last_hash == current_hash:  # param not changed
                    return None, self.param_info
                else:
                    return self.parameters, self.param_info
            else:
                return self.parameters, self.param_info
        else:
            raise ValueError('invalid request: '+str(request))


class ParameterClient(object):
    """
        On agent side, sends requests to parameter servers to fetch the
        latest parameters.
    """

    def __init__(self, host, port, timeout=2):
        """
        Args:
            host: parameter server host
            port: parameter server port
            timeout: how long should the the client wait
                if the parameter server is not available
        """
        self.host = host
        self.port = port
        self.timeout = timeout
        self._current_info = {}
        self._last_hash = ''
        self.alive = False

        self._client = ZmqClient(
            host=self.host,
            port=self.port,
            timeout=self.timeout,
            serializer=U.serialize,
            deserializer=U.deserialize)

    def fetch_parameter_with_info(self, force_update=False):
        """
            Called by agent to retrieve parameters
            By default, pulls from PS ONLY WHEN the parameter hash changes
                to prevent duplicate fetching. No-op when duplicate.
            Caching can be overriden by force_update

        Args:
            force_update: forces download of parameter, regardless of
                currently cached hash

        Returns:
            (param or None, info or None)
        """
        try:
            if force_update:
                response = self._client.request('parameter')
            else:
                response = self._client.request('parameter:' + self._last_hash)
        except ZmqTimeoutError:
            self.on_fetch_parameter_failed()
            return None, None
        self.on_fetch_parameter_success()
        param, info = response
        if info is None:
            return None, None

        self._last_hash = info['hash']
        return param, info

    def fetch_info(self):
        """
            Fetch the metadata of parameters on parameter server

        Returns:
            dictionary of metadata
        """
        try:
            response = self._client.request('info')
        except ZmqTimeoutError:
            self.on_fetch_parameter_failed()
            return None
        self.on_fetch_parameter_success()
        _, info = response
        return info

    def on_fetch_parameter_failed(self):
        """
            Called when connection with parameter server fails
            to be established
        """
        if self.alive:
            self.alive = False
            print('Parameter client request timed out')

    def on_fetch_parameter_success(self):
        """
            Called when connection with parameter server
            is succesfully established
        """
        if not self.alive:
            self.alive = True
            print('Parameter client came back alive')



"""
Actor function
"""
import torch
import random
from torch.autograd import Variable
import surreal.utils as U
from surreal.model.q_net import build_ffqfunc
from .base import Agent


class QAgent(Agent):
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 agent_id,
                 agent_mode,
                 render=False):
        super().__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            agent_id=agent_id,
            agent_mode=agent_mode,
            render=render,
        )
        self.q_func, self.action_dim = build_ffqfunc(
            self.learner_config,
            self.env_config
        )
        self.exploration = self.get_exploration_schedule()
        self.T = 0

    def act(self, obs):
        assert torch.is_tensor(obs)
        if self.agent_mode == 'training':
            eps = self.exploration.value(self.T)
        else:
            eps = self.learner_config.eval.eps
        self.T += 1
        if self.agent_mode == 'eval_deterministic' or random.random() > eps:
            obs = obs[None]  # vectorize
            obs = Variable(obs, volatile=True)
            q_values = self.q_func(obs)
            return U.to_scalar(q_values.data.max(1)[1].cpu())
        else:  # random exploration
            return random.randrange(self.action_dim)

    def module_dict(self):
        return {
            'q_func': self.q_func
        }

    def default_config(self):
        return {
            'model': {
                'convs': '_list_',
                'fc_hidden_sizes': '_list_',
                'dueling': '_bool_'
            },
            'eval': {
                'eps': '_float_'
            }
        }

    def get_exploration_schedule(self):
        C = self.learner_config.algo.exploration
        if C.schedule.lower() == 'linear':
            return U.LinearSchedule(
                initial_p=1.0,
                final_p=C.final_eps,
                schedule_timesteps=int(C.steps),
            )
        else:
            steps = C.steps
            final_epses = C.final_eps
            U.assert_type(steps, list)
            U.assert_type(final_epses, list)
            assert len(steps) == len(final_epses)
            endpoints = [(0, 1.0)]
            for step, eps in zip(steps, final_epses):
                endpoints.append((step, eps))
            return U.PiecewiseSchedule(
                endpoints=endpoints,
                outside_value=final_epses[-1]
            )




from .exp_sender import ExpSender
from .exp_collector import ExperienceCollectorServer
from .data_fetcher import LearnerDataPrefetcher
from .module_dict import ModuleDict
from .parameter_server import (
    ParameterPublisher,
    ParameterClient,
    ShardedParameterServer,
    ParameterPublisher,
    )


import torch
import torch.nn as nn
import numpy as np
from .base import Learner
from .aggregator import SSARAggregator, FrameStackPreprocessor
from surreal.model.ddpg_net import DDPGModel
from surreal.session import BASE_LEARNER_CONFIG, ConfigError
import surreal.utils as U
import torchx as tx


class DDPGLearner(Learner):
    '''
    DDPGLearner: subclass of Learner that contains DDPG algorithm logic
    Attributes:
        gpu_option: 'cpu' if not using GPU, 'cuda:all' otherwise
        model: instance of DDPGModel from surreal.model.ddpg_net
        model_target: instance of DDPGModel, used as a reference policy
            for Bellman updates
        use_action_regularization: boolean flag -- regularization method based on
            https://arxiv.org/pdf/1802.09477.pdf
        use_double_critic: boolean flag -- overestimation bias correction based on
            https://arxiv.org/pdf/1802.09477.pdf
        [actor/critic]_optim: Adam Optimizer for policy and baseline network
        aggregator: experience aggregator used to batch experiences from
            a list of experiences into a format usable by the model.
            For available aggregators, see surreal.learner.aggregator
        target_update_type: 'hard' update sets the weights of model_target equal to model
            after target_update_interval steps, whereas 'soft' update moves the parameters
            of model_target towards model after every step
        total_learn_time, forward_time, etc: timers that measure average time spent in
            each operation of the learner. These timers will be reported in tensorboard.

    important member functions:
        private methods:
        _optimize: function that makes policy and value function updates

        public methods:
        learn: method to perform optimization and send to tensorplex for log
        module_dict: returns the corresponding parameters
        preprocess: this function is called in learner/main prior to learn(),
            This operation occurs in a separate thread, meaning that conversion
            from numpy arrays to gpu tensors can occur asynchronously to gpu
            processing operations in learn().

    Arguments:
        learner_config, env_config, session_config: experiment setup configurations.  An example set of configs
            can be found at surreal/main/ddpg_configs.py.  Note that the surreal/env/make_env function adds attributes
            env_config.action_spec and env_config.obs_spec, which are required for this init method to function
            properly.
    '''

    def __init__(self, learner_config, env_config, session_config):
        super().__init__(learner_config, env_config, session_config)

        self.current_iteration = 0

        # load multiple optimization instances onto a single gpu
        self.batch_size = self.learner_config.replay.batch_size
        self.discount_factor = self.learner_config.algo.gamma
        self.n_step = self.learner_config.algo.n_step
        self.is_pixel_input = self.env_config.pixel_input
        self.use_layernorm = self.learner_config.model.use_layernorm
        self.use_double_critic = self.learner_config.algo.network.use_double_critic
        self.use_action_regularization = self.learner_config.algo.network.use_action_regularization

        self.frame_stack_concatenate_on_env = self.env_config.frame_stack_concatenate_on_env

        self.log.info('Initializing DDPG learner')
        self._num_gpus = session_config.learner.num_gpus
        if not torch.cuda.is_available():
            self.gpu_ids = 'cpu'
            self.log.info('Using CPU')
        else:
            self.gpu_ids = 'cuda:all'
            self.log.info('Using GPU')
            self.log.info('cudnn version: {}'.format(torch.backends.cudnn.version()))
            torch.backends.cudnn.benchmark = True
            self._num_gpus = 1

        with tx.device_scope(self.gpu_ids):
            self._target_update_init()

            self.clip_actor_gradient = self.learner_config.algo.network.clip_actor_gradient
            if self.clip_actor_gradient:
                self.actor_gradient_clip_value = self.learner_config.algo.network.actor_gradient_value_clip
                self.log.info('Clipping actor gradient at {}'.format(self.actor_gradient_clip_value))

            self.clip_critic_gradient = self.learner_config.algo.network.clip_critic_gradient
            if self.clip_critic_gradient:
                self.critic_gradient_clip_value = self.learner_config.algo.network.critic_gradient_value_clip
                self.log.info('Clipping critic gradient at {}'.format(self.critic_gradient_clip_value))

            self.action_dim = self.env_config.action_spec.dim[0]
            self.model = DDPGModel(
                obs_spec=self.env_config.obs_spec,
                action_dim=self.action_dim,
                use_layernorm=self.use_layernorm,
                actor_fc_hidden_sizes=self.learner_config.model.actor_fc_hidden_sizes,
                critic_fc_hidden_sizes=self.learner_config.model.critic_fc_hidden_sizes,
                conv_out_channels=self.learner_config.model.conv_spec.out_channels,
                conv_kernel_sizes=self.learner_config.model.conv_spec.kernel_sizes,
                conv_strides=self.learner_config.model.conv_spec.strides,
                conv_hidden_dim=self.learner_config.model.conv_spec.hidden_output_dim,
            )

            self.model_target = DDPGModel(
                obs_spec=self.env_config.obs_spec,
                action_dim=self.action_dim,
                use_layernorm=self.use_layernorm,
                actor_fc_hidden_sizes=self.learner_config.model.actor_fc_hidden_sizes,
                critic_fc_hidden_sizes=self.learner_config.model.critic_fc_hidden_sizes,
                conv_out_channels=self.learner_config.model.conv_spec.out_channels,
                conv_kernel_sizes=self.learner_config.model.conv_spec.kernel_sizes,
                conv_strides=self.learner_config.model.conv_spec.strides,
                conv_hidden_dim=self.learner_config.model.conv_spec.hidden_output_dim,
            )

            if self.use_double_critic:
                self.model2 = DDPGModel(
                    obs_spec=self.env_config.obs_spec,
                    action_dim=self.action_dim,
                    use_layernorm=self.use_layernorm,
                    actor_fc_hidden_sizes=self.learner_config.model.actor_fc_hidden_sizes,
                    critic_fc_hidden_sizes=self.learner_config.model.critic_fc_hidden_sizes,
                    conv_out_channels=self.learner_config.model.conv_spec.out_channels,
                    conv_kernel_sizes=self.learner_config.model.conv_spec.kernel_sizes,
                    conv_strides=self.learner_config.model.conv_spec.strides,
                    conv_hidden_dim=self.learner_config.model.conv_spec.hidden_output_dim,
                    critic_only=True,
                )

                self.model_target2 = DDPGModel(
                    obs_spec=self.env_config.obs_spec,
                    action_dim=self.action_dim,
                    use_layernorm=self.use_layernorm,
                    actor_fc_hidden_sizes=self.learner_config.model.actor_fc_hidden_sizes,
                    critic_fc_hidden_sizes=self.learner_config.model.critic_fc_hidden_sizes,
                    conv_out_channels=self.learner_config.model.conv_spec.out_channels,
                    conv_kernel_sizes=self.learner_config.model.conv_spec.kernel_sizes,
                    conv_strides=self.learner_config.model.conv_spec.strides,
                    conv_hidden_dim=self.learner_config.model.conv_spec.hidden_output_dim,
                    critic_only=True,
                )

            self.critic_criterion = nn.MSELoss()

            self.log.info('Using Adam for critic with learning rate {}'.format(self.learner_config.algo.network.lr_critic))
            self.critic_optim = torch.optim.Adam(
                self.model.get_critic_parameters(),
                lr=self.learner_config.algo.network.lr_critic,
                weight_decay=self.learner_config.algo.network.critic_regularization # Weight regularization term
            )

            self.log.info('Using Adam for actor with learning rate {}'.format(self.learner_config.algo.network.lr_actor))
            self.actor_optim = torch.optim.Adam(
                self.model.get_actor_parameters(),
                lr=self.learner_config.algo.network.lr_actor,
                weight_decay=self.learner_config.algo.network.actor_regularization # Weight regularization term
            )

            if self.use_double_critic:
                self.log.info('Using Adam for critic with learning rate {}'.format(self.learner_config.algo.network.lr_critic))
                self.critic_optim2 = torch.optim.Adam(
                    self.model2.get_critic_parameters(),
                    lr=self.learner_config.algo.network.lr_critic,
                    weight_decay=self.learner_config.algo.network.critic_regularization # Weight regularization term
                )

            self.log.info('Using {}-step bootstrapped return'.format(self.learner_config.algo.n_step))
            self.frame_stack_preprocess = FrameStackPreprocessor(self.env_config.frame_stacks)
            self.aggregator = SSARAggregator(self.env_config.obs_spec, self.env_config.action_spec)

            self.model_target.actor.hard_update(self.model.actor)
            self.model_target.critic.hard_update(self.model.critic)

            if self.use_double_critic:
                self.model_target2.critic.hard_update(self.model2.critic)

            self.total_learn_time = U.TimeRecorder()
            self.forward_time = U.TimeRecorder()
            self.critic_update_time = U.TimeRecorder()
            self.actor_update_time = U.TimeRecorder()

    # override
    def preprocess(self, batch):
        '''
        Override for learner/base/preprocess.  Before learn() is called, preprocess() takes the batch and converts
        the numpy arrays to pytorch tensors.  Note that this operation will transfer the data to gpu if a gpu is used.

        Arguments:
            batch: a batch of numpy arrays from the replay memory
        '''
        # Convert all numpy arrays to pytorch tensors, and transfers to gpu if applicable
        with tx.device_scope(self.gpu_ids):
            obs, actions, rewards, obs_next, done = (
                batch['obs'],
                batch['actions'],
                batch['rewards'],
                batch['obs_next'],
                batch['dones']
            )
            device_name = 'cpu'
            if self._num_gpus > 0:
                device_name = 'cuda'

            for modality in obs:
                for key in obs[modality]:
                    if modality == 'pixel':
                        obs[modality][key] = (torch.tensor(obs[modality][key], dtype=torch.uint8)
                            .to(torch.device(device_name))).float().detach()
                    else:
                        obs[modality][key] = (torch.tensor(obs[modality][key], dtype=torch.float32)
                            .to(torch.device(device_name))).detach()

            for modality in obs_next:
                for key in obs_next[modality]:
                    if modality == 'pixel':
                        obs_next[modality][key] = (torch.tensor(obs_next[modality][key], dtype=torch.uint8)
                            .to(torch.device(device_name))).float().detach()
                    else:
                        obs_next[modality][key] = (torch.tensor(obs_next[modality][key], dtype=torch.float32)
                            .to(torch.device(device_name))).detach()

            actions = torch.tensor(actions, dtype=torch.float32).to(torch.device(device_name))
            rewards = torch.tensor(rewards, dtype=torch.float32).to(torch.device(device_name))
            done = torch.tensor(done, dtype=torch.float32).to(torch.device(device_name))

            (
                batch['obs'],
                batch['actions'],
                batch['rewards'],
                batch['obs_next'],
                batch['dones']
            ) = (
                obs,
                actions,
                rewards,
                obs_next,
                done
            )
            return batch

    def _optimize(self, obs, actions, rewards, obs_next, done):
        '''
        Note that while the replay contains uint8, the
        aggregator returns float32 tensors

        Arguments:
            obs: an observation from the minibatch, often represented as s_n in literature. Dimensionality: (N, C) for
                low dimensional inputs, (N, C, H, W) for pixel inputs
            actions: actions taken given observations obs, often represented as a_n in literature.
                Dimensionality: (N, A), where A is the dimensionality of a single action
            rewards: rewards received after action is taken. Dimensionality: N
            obs_next: an observation from the minibatch, often represented as s_{n+1} in literature
            done: 1 if obs_next is terminal, 0 otherwise. Dimensionality: N
        '''
        with tx.device_scope(self.gpu_ids):

            with self.forward_time.time():
                assert actions.max().item() <= 1.0
                assert actions.min().item() >= -1.0

                # estimate rewards using the next state: r + argmax_a Q'(s_{t+1}, u'(a))

                model_policy, next_Q_target = self.model_target.forward(obs_next)
                if self.use_action_regularization:
                    # https://github.com/sfujim/TD3/blob/master/TD3.py -- action regularization
                    policy_noise = 0.2
                    noise_clip = 0.5
                    batch_size = self.batch_size
                    noise = np.clip(np.random.normal(0, policy_noise, size=(batch_size, self.action_dim)), -noise_clip,
                                    noise_clip)
                    device_name = 'cpu'
                    if self._num_gpus > 0:
                        device_name = 'cuda'
                    model_policy += torch.tensor(noise, dtype=torch.float32).to(device_name).detach()
                    model_policy = model_policy.clamp(-1, 1).to(device_name)
                y = rewards + pow(self.discount_factor, self.n_step) * next_Q_target * (1.0 - done)
                if self.use_double_critic:
                    _, next_Q_target2 = self.model_target2.forward(obs_next, action=model_policy)
                    y2 = rewards + pow(self.discount_factor, self.n_step) * next_Q_target2 * (1.0 - done)
                    y = torch.min(y, y2)
                y = y.detach()

                # compute Q(s_t, a_t)
                perception = self.model.forward_perception(obs)
                y_policy = self.model.forward_critic(
                    perception,
                    actions.detach()
                )

                y_policy2 = None
                if self.use_double_critic:
                    perception2 = self.model2.forward_perception(obs)
                    y_policy2 = self.model2.forward_critic(
                        perception2,
                        actions.detach()
                    )

            # critic update
            with self.critic_update_time.time():
                self.model.critic.zero_grad()
                if self.is_pixel_input:
                    self.model.perception.zero_grad()
                critic_loss = self.critic_criterion(y_policy, y)        
                critic_loss.backward()
                if self.clip_critic_gradient:
                    self.model.critic.clip_grad_value(self.critic_gradient_clip_value)
                self.critic_optim.step()

                if self.use_double_critic:
                    self.model2.critic.zero_grad()
                    if self.is_pixel_input:
                        self.model2.perception.zero_grad()
                    critic_loss = self.critic_criterion(y_policy2, y)
                    critic_loss.backward()
                    if self.clip_critic_gradient:
                        self.model2.critic.clip_grad_value(self.critic_gradient_clip_value)
                    self.critic_optim2.step()

            # actor update
            with self.actor_update_time.time():
                self.model.actor.zero_grad()
                actor_loss = -self.model.forward_critic(
                    perception.detach(),
                    self.model.forward_actor(perception.detach())
                )
                actor_loss = actor_loss.mean()
                actor_loss.backward()
                if self.clip_actor_gradient:
                    self.model.actor.clip_grad_value(self.actor_gradient_clip_value)
                self.actor_optim.step()

            tensorplex_update_dict = {
                'actor_loss': actor_loss.item(),
                'critic_loss': critic_loss.item(),
                'action_norm': actions.norm(2, 1).mean().item(),
                'rewards': rewards.mean().item(),
                'Q_target': y.mean().item(),
                'Q_policy': y_policy.mean().item(),
                'performance/forward_time': self.forward_time.avg,
                'performance/critic_update_time': self.critic_update_time.avg,
                'performance/actor_update_time': self.actor_update_time.avg,
            }
            if self.use_double_critic:
                tensorplex_update_dict['Q_policy2'] = y_policy2.mean().item()

            # (possibly) update target networks
            self._target_update()

            return tensorplex_update_dict

    def learn(self, batch):
        '''
        Performs a gradient descent step on 'batch'

        Arguments:
            batch: a minibatch sampled from the replay memory, after preprocessing steps such as transfer to pytorch
            tensors and aggregation step
        '''
        self.current_iteration += 1
        with self.total_learn_time.time():
            tensorplex_update_dict = self._optimize(
                batch.obs,
                batch.actions,
                batch.rewards,
                batch.obs_next,
                batch.dones
            )
            tensorplex_update_dict['performance/total_learn_time'] = self.total_learn_time.avg
            self.tensorplex.add_scalars(tensorplex_update_dict, global_step=self.current_iteration)
            self.periodic_checkpoint(
                global_steps=self.current_iteration,
                score=None,
            )

    def module_dict(self):
        return {
            'ddpg': self.model,
        }

    def checkpoint_attributes(self):
        return [
            'current_iteration',
            'model', 'model_target'
        ]

    def _target_update_init(self):
        target_update_config = self.learner_config.algo.network.target_update
        self.target_update_type = target_update_config.type

        if self.target_update_type == 'soft':
            self.target_update_tau = target_update_config.tau
            self.log.info('Using soft target update with tau = {}'.format(self.target_update_tau))
        elif self.target_update_type == 'hard':
            self.target_update_counter = 0
            self.target_update_interval = target_update_config.interval
            self.log.info('Using hard target update every {} steps'.format(self.target_update_interval))
        else:
            raise ConfigError('Unsupported ddpg update type: {}'.format(target_update_config.type))

    def _target_update(self):
        '''
        Perform update on target model networks.  This update is either 'soft', meaning the target model drifts towards
        the current model at a rate tau, or 'hard', meaning the target model performs a hard copy operation on the
        current model every target_update_interval steps.
        '''
        if self.target_update_type == 'soft':
            self.model_target.actor.soft_update(self.model.actor, self.target_update_tau)
            self.model_target.critic.soft_update(self.model.critic, self.target_update_tau)
            if self.use_double_critic:
                self.model_target2.critic.soft_update(self.model2.critic, self.target_update_tau)
                if self.is_pixel_input:
                    self.model_target2.perception.soft_update(self.model2.perception, self.target_update_tau)
            if self.is_pixel_input:
                self.model_target.perception.soft_update(self.model.perception, self.target_update_tau)
        elif self.target_update_type == 'hard':
            self.target_update_counter += 1
            if self.target_update_counter % self.target_update_interval == 0:
                self.model_target.actor.hard_update(self.model.actor)
                self.model_target.critic.hard_update(self.model.critic)
                if self.use_double_critic:
                    self.model_target2.critic.hard_update(self.model2.critic)
                    if self.is_pixel_input:
                        self.model_target2.perception.hard_update(self.model2.perception)
                if self.is_pixel_input:
                    self.model_target.perception.hard_update(self.model.perception)

    # override
    def _prefetcher_preprocess(self, batch):
        '''
        If frame_stack_preprocess is not set, each experience in the replay will be stored as a list of frames, as
        opposed to a single numpy array.  We must condense them into a single numpy array as that is what the
        aggregator expects.
        '''
        if not self.frame_stack_concatenate_on_env:
            batch = self.frame_stack_preprocess.preprocess_list(batch)
        batch = self.aggregator.aggregate(batch)
        return batch



import weakref
from threading import Thread
import surreal.utils as U
from caraml.zmq import ZmqReceiver


class ExperienceCollectorServer(Thread):
    """
        Accepts experience from agents,
        deduplicates experience whenever possible
    """
    def __init__(self, host, port, exp_handler, load_balanced=True):
        Thread.__init__(self)
        self.host = host
        self.port = port
        self.load_balanced = load_balanced
        self._exp_handler = exp_handler
        # To be initialized in run()
        self._weakref_map = None
        self.receiver = None

    def run(self):
        """
            Starts the server loop
        """
        self._weakref_map = weakref.WeakValueDictionary()
        self.receiver = ZmqReceiver(host=self.host,
                                    port=self.port,
                                    bind=not self.load_balanced,
                                    deserializer=U.deserialize)
        while True:
            exp, storage = self.receiver.recv()
            experience_list = self._retrieve_storage(exp, storage)
            for exp in experience_list:
                self._exp_handler(exp)

    def _retrieve_storage(self, exp, storage):
        """
        Args:
            exp: a nested dict or list
                Only dict keys that end with `_hash` will be retrieved.
                The processed key will see `_hash` removed
            storage: chunk of storage sent with the exps
        """
        if isinstance(exp, list):
            for i, e in enumerate(exp):
                exp[i] = self._retrieve_storage(e, storage)

        elif isinstance(exp, dict):
            for key in list(exp.keys()):  # copy keys
                if key.endswith('_hash'):
                    new_key = key[:-len('_hash')]  # delete suffix
                    exp[new_key] = self._retrieve_storage(exp[key], storage)
                    del exp[key]
                else:
                    exp[key] = self._retrieve_storage(exp[key], storage)

        elif isinstance(exp, str):
            exphash = exp
            if exphash in self._weakref_map:
                return self._weakref_map[exphash]
            else:
                self._weakref_map[exphash] = storage[exphash]
                return storage[exphash]
        return exp



"""
A template class that defines base agent APIs
"""
import time
import os
import surreal.utils as U
from surreal.env import make_env
from surreal.session import (
    PeriodicTracker, PeriodicTensorplex,
    get_loggerplex_client, get_tensorplex_client,
)
from surreal.distributed import ParameterClient, ModuleDict
from surreal.env import (
    MaxStepWrapper,
    TrainingTensorplexMonitor,
    EvalTensorplexMonitor,
    VideoWrapper
)

AGENT_MODES = ['training', 'eval_deterministic', 'eval_stochastic']


class Agent(object, metaclass=U.AutoInitializeMeta):
    """
        Important: When extending this class, make sure to follow the init method signature so that 
        orchestrating functions can properly initialize custom agents.

        TODO: Extend the initilization to allow custom non-config per-agent settings.
            To be used to have a heterogeneous agent population
    """
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 agent_id,
                 agent_mode,
                 render=False):
        """
            Initialize the agent class, 
        """
        self.learner_config = learner_config
        self.env_config = env_config
        self.session_config = session_config

        assert agent_mode in AGENT_MODES
        self.agent_mode = agent_mode
        self.agent_id = agent_id

        self._setup_parameter_pull()
        self._setup_logging()

        self.current_episode = 0
        self.cumulative_steps = 0
        self.current_step = 0

        self.actions_since_param_update = 0
        self.episodes_since_param_update = 0

        self.render = render

    #######
    # Internal initialization methods
    #######
    def _initialize(self):
        """
            implements AutoInitializeMeta meta class.
            self.module_dict can only happen after the module is constructed by subclasses.
        """
        host, port = os.environ['SYMPH_PS_FRONTEND_HOST'], os.environ['SYMPH_PS_FRONTEND_PORT']
        self._module_dict = self.module_dict()
        if not isinstance(self._module_dict, ModuleDict):
            self._module_dict = ModuleDict(self._module_dict)
        self._ps_client = ParameterClient(
            host=host,
            port=port,
        )

    def _setup_parameter_pull(self):
        self._fetch_parameter_mode = self.session_config.agent.fetch_parameter_mode
        self._fetch_parameter_interval = self.session_config.agent.fetch_parameter_interval
        self._fetch_parameter_tracker = PeriodicTracker(self._fetch_parameter_interval)

    def _setup_logging(self):
        """
            Creates tensorplex logger and loggerplex logger
            Initializes bookkeeping values
        """
        if self.agent_mode == 'training':
            logger_name = 'agent-{}'.format(self.agent_id)
            self.tensorplex = self._get_tensorplex(
                '{}/{}'.format('agent', self.agent_id))
        else:
            logger_name = 'eval-{}'.format(self.agent_id)
            self.tensorplex = self._get_tensorplex(
                '{}/{}'.format('eval', self.agent_id))

        self.log = get_loggerplex_client(logger_name, self.session_config)
        # record how long the current parameter have been used
        self.actions_since_param_update = 0
        self.episodes_since_param_update = 0
        # Weighted Average over ~100 parameter updates.
        self.actions_per_param_update = U.MovingAverageRecorder(decay=0.99)
        self.episodes_per_param_update = U.MovingAverageRecorder(decay=0.99)

    def _get_tensorplex(self, name):
        """
            Get the periodic tensorplex object
        Args:
            @name: The name of the collection of metrics
        """
        tp = get_tensorplex_client(
            name,
            self.session_config
        )
        periodic_tp = PeriodicTensorplex(
            tensorplex=tp,
            period=self.session_config.tensorplex.update_schedule.agent,
            is_average=True,
            keep_full_history=False
        )
        return periodic_tp

    #######
    # Exposed abstract methods
    # Override in subclass, no need to call super().act etc.
    # Enough for basic usage
    #######
    def act(self, obs):
        """
        Abstract method for taking actions.
        You should check `self.agent_mode` in the function and change act()
        logic with respect to training VS evaluation.

        Args:
            obs: typically a single obs, make sure to vectorize it first before
                passing to the torch `model`.

        Returns:
            action to be executed in the env
        """
        raise NotImplementedError

    def module_dict(self):
        """
        Returns:
            a dict of name -> surreal.utils.pytorch.Module
        """
        raise NotImplementedError

    #######
    # Advanced exposed methods
    # Override in subclass, NEED to call super().on_parameter_fetched() etc.
    # User need to take care of agent mode
    # For advanced usage
    #######
    def on_parameter_fetched(self, params, info):
        """
            Called when a new parameter is fetched.
        """
        if self.agent_mode == 'training':
            # The time it takes for parameter to go from learner to agent
            delay = time.time() - info['time']
            self.actions_per_param_update.add_value(
                self.actions_since_param_update)
            self.episodes_per_param_update.add_value(
                self.episodes_since_param_update)
            self.tensorplex.add_scalars(
                {
                    '.core/parameter_publish_delay_s': delay,
                    '.core/actions_per_param_update':
                        self.actions_per_param_update.cur_value(),
                    '.core/episodes_per_param_update':
                        self.episodes_per_param_update.cur_value()
                })
            self.actions_since_param_update = 0
            self.episodes_since_param_update = 0
        return params

    def pre_action(self, obs):
        """
            Called before act is called by agent main script
        """
        if self.agent_mode == 'training':
            if self._fetch_parameter_mode == 'step' and \
                    self._fetch_parameter_tracker.track_increment():
                self.fetch_parameter()

    def post_action(self, obs, action, obs_next, reward, done, info):
        """
            Called after act is called by agent main script
        """
        self.current_step += 1
        self.cumulative_steps += 1
        if self.agent_mode == 'training':
            self.actions_since_param_update += 1
            if done:
                self.episodes_since_param_update += 1

    def pre_episode(self):
        """
            Called by agent process.
            Can beused to reset internal states before an episode starts
        """
        if self.agent_mode == 'training':
            if self._fetch_parameter_mode == 'episode' and \
                    self._fetch_parameter_tracker.track_increment():
                self.fetch_parameter()

    def post_episode(self):
        """
            Called by agent process.
            Can beused to reset internal states after an episode ends
            I.e. after the post_action when done = True
        """
        self.current_episode += 1

    #######
    # Main loops.
    # Customize this to fully customize the agent process
    #######
    def main(self):
        """
            Default Main loop
        Args:
            @env: the environment to run agent on
        """
        self.main_setup()
        while True:
            self.main_loop()

    def main_setup(self):
        """
            Setup before constant looping
        """
        env = self.get_env()
        env = self.prepare_env(env)
        self.env = env
        self.fetch_parameter()

    def main_loop(self):
        """
            One loop of agent, runs one episode of the environment
        """
        env = self.env
        self.pre_episode()
        obs, info = env.reset()
        total_reward = 0.0
        while True:
            if self.render:
                env.render()
            self.pre_action(obs)
            action = self.act(obs)
            obs_next, reward, done, info = env.step(action)
            total_reward += reward
            self.post_action(obs, action, obs_next, reward, done, info)
            obs = obs_next
            if done:
                break
        self.post_episode()
        if self.current_episode % 20 == 0:
            self.log.info('Episode {} reward {}'
                          .format(self.current_episode,
                                  total_reward))

    def get_env(self):
        """
        Returns a subclass of EnvBase, created from self.env_config
        """
        if self.agent_mode in ['eval_deterministic', 'eval_stochastic']:
            env, _ = make_env(self.env_config, mode='eval')
        else:
            env, _ = make_env(self.env_config)
        return env

    def prepare_env(self, env):
        """
            Applies custom wrapper to the environment as necessary
        Args:
            @env: subclass of EnvBse

        Returns:
            @env: The (possibly wrapped) environment
        """
        if self.agent_mode == 'training':
            return self.prepare_env_agent(env)
        else:
            return self.prepare_env_eval(env)

    def prepare_env_agent(self, env):
        """
            Applies custom wrapper to the environment as necessary
            Only changes agent behavior
        """
        # This has to go first as it alters step() return value
        limit_episode_length = self.env_config.limit_episode_length
        if limit_episode_length > 0:
            env = MaxStepWrapper(env, limit_episode_length)
        env = TrainingTensorplexMonitor(
            env,
            agent_id=self.agent_id,
            session_config=self.session_config,
            separate_plots=True
        )
        return env

    def prepare_env_eval(self, env):
        """
            Applies custom wrapper to the environment as necessary
            Only changes eval behavior
        """
        limit_episode_length = self.env_config.limit_episode_length
        if limit_episode_length > 0:
            env = MaxStepWrapper(env, limit_episode_length)

        env = EvalTensorplexMonitor(
            env,
            eval_id=self.agent_id,
            fetch_parameter=self.fetch_parameter,
            session_config=self.session_config,
        )

        env_category = self.env_config.env_name.split(':')[0]
        if self.env_config.video.record_video and self.agent_id == 0:
            if env_category != 'gym':
                # gym video recording not supported due to bug in OpenAI gym
                # https://github.com/openai/gym/issues/1050
                env = VideoWrapper(env, self.env_config, self.session_config)
        return env

    def main_agent(self):
        """
            Main loop ran by the agent script
            Override if you want to customize agent behavior completely
        """
        self.main()

    def main_eval(self):
        """
            Main loop ran by the eval script
            Override if you want to customize eval behavior completely
        """
        self.main()

    #######
    # Exposed public methods
    #######
    def fetch_parameter(self):
        """
            Extends base class fetch_parameters to add some logging
        """
        params, info = self._ps_client.fetch_parameter_with_info()
        if params:
            params = U.deserialize(params)
            params = self.on_parameter_fetched(params, info)
            self._module_dict.load(params)

    def fetch_parameter_info(self):
        """
            Fetch information about the parameters currently held by the parameter server
        """
        return self._ps_client.fetch_info()

    def set_agent_mode(self, agent_mode):
        """
        Args:
            agent_mode: 'training', 'eval_deterministic', or 'eval_stochastic'
        """
        assert agent_mode in AGENT_MODES
        self.agent_mode = agent_mode



from .config import extend_config

# ======================== Agent-Learner side ========================
BASE_LEARNER_CONFIG = {
    'model': '_dict_',
    'algo': {
        # Agent class to instantiate
        # Learner class to instantiate
        'n_step': 1,
        'gamma': '_float_',
        'use_batchnorm': False,
        'limit_training_episode_length': 0,
        'network': {
            'actor_regularization': 0.0,
            'critic_regularization': 0.0,
        },
    },
    'replay': {
        # The replay class to instantiate
        'batch_size': '_int_',
        'replay_shards': 1,
    },
    'parameter_publish': {
        # Minimum amount of time (seconds) between two parameter publish
        'min_publish_interval': 0.3, 
    },
}


# ======================== Env side ========================
BASE_ENV_CONFIG = {
    'env_name' : '_str_',
    'sleep_time': 0.0,
    'video' : {
        'record_video' : False,
        'max_videos': 10,
        'record_every': 10,
        'save_folder': None,
    },
    'eval_mode': {}, # for providing different env init args when in eval
    'action_spec': {},
    'obs_spec': {},
    'frame_stacks': 1,
    'frame_stack_concatenate_on_env': True,
    # 'action_spec': {
    #     'dim': '_list_',
    #     'type': '_enum[continuous, discrete]_'
    # },
    # 'obs_spec': {
    #     'dim': '_list_',
    #     'type': ''  # TODO uint8 format
    # },
}


# ======================== Session side ========================
BASE_SESSION_CONFIG = {
    'folder': '_str_',

    'replay': {
        'collector_frontend_host': '_str_',  # upstream from agents' pusher
        'collector_frontend_port': '_int_',
        'collector_backend_host': '_str_',  # upstream from agents' pusher
        'collector_backend_port': '_int_',
        'sampler_frontend_host': '_str_',  # downstream to Learner request
        'sampler_frontend_port': '_int_',
        'sampler_backend_host': '_str_',  # downstream to Learner request
        'sampler_backend_port': '_int_',
        'max_puller_queue': '_int_',  # replay side: pull queue size
        'evict_interval': '_float_',  # in seconds
        'tensorboard_display': True,  # display replay stats on Tensorboard
    },
    'sender': {
        'flush_iteration': '_int_',
        'flush_time': '_int_',
    },
    'ps': {
        'parameter_serving_frontend_host': '_str_',
        'parameter_serving_frontend_port': '_int_',
        'parameter_serving_backend_host': '_str_',
        'parameter_serving_backend_port': '_int_',
        'shards': '_int_',
        'publish_host': '_str',  # upstream from learner
        'publish_port': '_int_'
    },
    'tensorplex': {
        'host': '_str_',
        'port': '_int_',
        'tensorboard_port': '_int_',  # tensorboard port
        'agent_bin_size': 8,
        'max_processes': 4,
        'update_schedule': {  # TODO rename this to 'periodic'
            # for TensorplexWrapper:
            'training_env': '_int_',  # env record every N episodes
            'eval_env': '_int_',
            'eval_env_sleep': '_int_',  # throttle eval by sleep n seconds
            # for manual updates:
            'agent': '_int_',  # agent.tensorplex.add_scalars()
            # WARN!!: DEPRECATED
            'learner': '_int_',  # learner.tensorplex.add_scalars()
            'learner_min_update_interval': '_int_', #Update tensorplex at most every ? seconds
        }
    },
    'loggerplex': {
        'host': '_str_',
        'port': '_int_',
        'overwrite': False,
        'level': 'info',
        'show_level': True,
        'time_format': 'hms',
        'enable_local_logger': '_bool_',
        'local_logger_level': 'info',
        'local_logger_time_format': 'hms'
    },
    'agent': {
        'fetch_parameter_mode': '_str_',
        'fetch_parameter_interval': int,
    },
    'learner': {
        'num_gpus': '_int_',
        'prefetch_host': '_str_',
        'prefetch_port': '_int_',
        'prefetch_processes': '_int_',
        'max_prefetch_queue': '_int_',  # learner side: max number of batches to prefetch
        'max_preprocess_queue': '_int_',  # learner side: max number of batches to preprocess
    },
    'checkpoint': {
        'restore': '_bool_',  # if False, ignore the other configs under 'restore'
        'restore_folder': None,  # if None, use the same session folder.
                            # Otherwise restore ckpt from another experiment dir.
        'learner': {
            'restore_target': '_int_',
            'mode': '_enum[best,history]_',
            'keep_history': '_int_',
            'keep_best': '_int_',
            'periodic': '_int_',
            'min_interval': '_int_',
        },
        'agent': {
            'restore_target': '_int_',
            'mode': '_enum[best,history]_',
            'keep_history': '_int_',
            'keep_best': '_int_',
            'periodic': '_int_',
        },
    }
}


LOCAL_SESSION_CONFIG = {
    'folder': '_str_',

    'replay': {
        'collector_frontend_host': 'localhost',  # upstream from agents' pusher
        'collector_frontend_port': 7001,
        'collector_backend_host': 'localhost',  # upstream from agents' pusher
        'collector_backend_port': 7002,
        'sampler_frontend_host': 'localhost',  # downstream to Learner request
        'sampler_frontend_port': 7003,
        'sampler_backend_host': 'localhost',  # downstream to Learner request
        'sampler_backend_port': 7004,
        'max_puller_queue': 10000,  # replay side: pull queue size
        'evict_interval': 0.,  # in seconds
        'tensorboard_display': True,  # display replay stats on Tensorboard
    },
    'sender': {
        'flush_iteration': '_int_',
        'flush_time': 0,
    },
    'ps': {
        'parameter_serving_frontend_host': 'localhost',
        'parameter_serving_frontend_port': 7005,
        'parameter_serving_backend_host': 'localhost',
        'parameter_serving_backend_port': 7006,
        'shards': 2,
        'publish_host': 'localhost',  # upstream from learner
        'publish_port': 7007
    },
    'tensorplex': {
        'host': 'localhost',
        'port': 7008,
        'tensorboard_port': 6006,
        'update_schedule': { # TODO: rename this to 'periodic'
            # for TensorplexWrapper:
            'training_env': 20,  # env record every N episodes
            'eval_env': 20,
            'eval_env_sleep': 30,  # throttle eval by sleep n seconds
            # for manual updates:
            'agent': 20,  # agent.tensorplex.add_scalars()
            'learner': 20,  # learner.tensorplex.add_scalars()
            'learner_min_update_interval': 30, #Update tensorplex at most every 30 seconds
        }
    },
    'loggerplex': {
        'host': 'localhost',
        'port': 7009,
        'enable_local_logger': True,
    },
    'agent': {
        # fetch_parameter_mode: 'episode', 'episode:<n>', 'step', 'step:<n>'
        # every episode, every n episodes, every step, every n steps
        'fetch_parameter_mode': 'episode',
        'fetch_parameter_interval': 1,
    },
    'learner': {
        'num_gpus': 0,
        'prefetch_host': 'localhost',
        'prefetch_port': 7010,
        'prefetch_processes': 2,
        'max_prefetch_queue': 10,  # learner side: max number of batches to prefetch
        'max_preprocess_queue': 2,  # learner side: max number of batches to preprocess
    },
    'checkpoint': {
        'restore': False,  # if False, ignore the other configs under 'restore'
        'restore_folder': None,
        'learner': {
            'restore_target': 0,
            'mode': 'history',
            'keep_history': 2,
            'keep_best': 0, # TODO don't keep best unless we solve the learner score issue
            'periodic': 100000, # Save every 100000 steps
            'min_interval': 15 * 60, # No checkpoint less than 15 min apart.
        },
        'agent': {
            'restore_target': 0,
            'mode': 'history',
            'keep_history': 2,
            'keep_best': 0, # TODO don't keep best unless we solve the learner score issue
            'periodic': 100,
        },
    }
}

LOCAL_SESSION_CONFIG = extend_config(LOCAL_SESSION_CONFIG, BASE_SESSION_CONFIG)


KUBE_SESSION_CONFIG = {
    'folder': '_str_',

    'replay': {
        'collector_frontend_host': '_str_',  # upstream from agents' pusher
        'sampler_frontend_host': '_str_',  # downstream to Learner request
    },
    'sender': {
        'flush_iteration': '_int_',
    },
    'ps': {
        'host': '_str_',  # downstream to agent requests
        'publish_host': '_str_',  # upstream from learner
    },
    'tensorplex': {
        'host': '_str_',
    },
    'loggerplex': {
        'host': '_str_',
    },
}

KUBE_SESSION_CONFIG = extend_config(KUBE_SESSION_CONFIG, LOCAL_SESSION_CONFIG)



ATARI_NAMES = [
    'air_raid',
    'alien',
    'amidar',
    'assault',
    'asterix',
    'asteroids',
    'atlantis',
    'bank_heist',
    'battle_zone',
    'beam_rider',
    'berzerk',
    'bowling',
    'boxing',
    'breakout',
    'carnival',
    'centipede',
    'chopper_command',
    'crazy_climber',
    'demon_attack',
    'double_dunk',
    'elevator_action',
    'enduro',
    'fishing_derby',
    'freeway',
    'frostbite',
    'gopher',
    'gravitar',
    'hero',
    'ice_hockey',
    'jamesbond',
    'journey_escape',
    'kangaroo',
    'krull',
    'kung_fu_master',
    'montezuma_revenge',
    'ms_pacman',
    'name_this_game',
    'phoenix',
    'pitfall',
    'pong',
    'pooyan',
    'private_eye',
    'qbert',
    'riverraid',
    'road_runner',
    'robotank',
    'seaquest',
    'skiing',
    'solaris',
    'space_invaders',
    'star_gunner',
    'tennis',
    'time_pilot',
    'tutankham',
    'up_n_down',
    'venture',
    'video_pinball',
    'wizard_of_wor',
    'yars_revenge',
    'zaxxon'
]


def _camelcase_names(names):
    return [''.join([s.capitalize() for s in name.split('_')])
            for name in names]


ATARI_NAMES_CAP = _camelcase_names(ATARI_NAMES)
_ATARI_CAP_MAP = {name.lower(): name for name in ATARI_NAMES_CAP}


def atari_name_cap(name):
    """
    Capitalize an Atari game name properly
    """
    name = name.lower()
    if name not in _ATARI_CAP_MAP:
        raise KeyError(name + ' is not a valid Atari game')
    return _ATARI_CAP_MAP[name]



import json
import yaml
import re
import surreal.utils as U


class ConfigError(Exception):
    pass


def _trace_key(dict_trace, key):
    return 'key "{}" '.format('.'.join(dict_trace + [key]))


def _check_enum(value, formula):
    """
    Specify required enum, syntax "_enum[option1, option2, ..]_"
    All options are case insensitive.
    """

_enum_marker = re.compile('_enum\[(.*)\]_')


def _req_type_check(value):
    if not isinstance(value, str):
        return None
    value = value.lower()
    if value == '_object_':
        return lambda x: True
    elif value == '_singleton_':
        return lambda x: not isinstance(x, (list, dict))  # non-collection
    elif value == '_list_':
        return lambda x: isinstance(x, list)
    elif value == '_dict_':
        return lambda x: isinstance(x, dict)
    elif value == '_int_':
        return lambda x: isinstance(x, int)
    elif value == '_float_':
        return lambda x: isinstance(x, float)
    elif value == '_num_':
        return lambda x: isinstance(x, (int, float))
    elif value == '_str_':
        return lambda x: isinstance(x, str)
    elif value == '_bool_':
        return lambda x: isinstance(x, bool)
    elif _enum_marker.match(value):
        enum_options = _enum_marker.match(value).group(1)
        if not enum_options:
            raise ConfigError('_enum[...]_ cannot be empty')
        enum_options = list(map(str.strip, enum_options.split(',')))
        return lambda x: x in enum_options
    else:
        return None


def _is_req(value):
    return _req_type_check(value) is not None


def _has_req(config):
    for key, val in config.items():
        if _is_req(val):
            return True
        elif isinstance(val, dict):
            if _has_req(val):
                return True
    return False


def _raise_req_error(key, value, dict_trace, prefix_msg=''):
    value = value.lower()
    if value == '_object_':
        msg = 'filled'
    elif value == '_singleton_':
        msg = 'a singleton (non-list/dict)'
    elif value == '_list_':
        msg = 'a list'
    elif value == '_dict_':
        msg = 'a dict'
    elif value == '_int_':
        msg = 'an integer'
    elif value == '_float_':
        msg = 'a float'
    elif value == '_num_':
        msg = 'a numeric value'
    elif value == '_str_':
        msg = 'a string'
    elif value == '_bool_':
        msg = 'a boolean'
    elif _enum_marker.match(value):
        enum_options = _enum_marker.match(value).group(1)
        msg = 'an enum in [{}]'.format(enum_options)
    else:
        assert False, 'internal error: req value fell through'
    prefix_msg = prefix_msg+' ' if prefix_msg else ''
    raise ConfigError('{}{}must be {}.'
                      .format(prefix_msg, _trace_key(dict_trace, key), msg))


def _fill_default_config(config, default_config, dict_trace):
    for key, default_value in default_config.items():
        if key not in config:
            if _is_req(default_value):
                _raise_req_error(key, default_value, dict_trace,
                                 'Required entry missing:')
            elif isinstance(default_value, dict):
                if _has_req(default_value):
                    raise ConfigError(
                        'Sub-dict under key "{}" contains a required config: {}.'
                        .format(_trace_key(dict_trace, key), default_value)
                    )
            config[key] = default_value
        else:
            value = config[key]
            if _is_req(default_value):
                # value remains a placeholder after being extended
                if _is_req(value):
                    if value != default_value:
                        raise ConfigError(
                            'inherited {}: "{}" must match default "{}"'
                            .format(_trace_key(dict_trace, key),
                                    value, default_value)
                        )
                else:
                    checker = _req_type_check(default_value)
                    if not checker(value):
                        _raise_req_error(
                            key, default_value, dict_trace, 'Wrong type:')
            else:
                if (isinstance(value, dict) and
                        not isinstance(default_value, dict)):
                    raise ConfigError(
                        _trace_key(dict_trace, key)
                        + 'must be a singleton instead of a sub-dict'
                    )
                if isinstance(default_value, dict):
                    if not isinstance(value, dict):
                        raise ConfigError(
                            _trace_key(dict_trace, key)
                            + 'must have a sub-dict instead of a singleton'
                        )
                    config[key] = _fill_default_config(
                        value,
                        default_value,
                        dict_trace + [key]
                    )
    return config


_RESERVED_METHODS = ['keys', 'items', 'values', 'get', 'copy', 'update',
                     'extend', 'load_file', 'dump_file', 'to_dict']


class Config(dict):
    """
    Source code modified from EasyDict to support extra methods.
    Does not support subclass attribute any more.
    https://github.com/makinacorpus/easydict
    """
    def __init__(self, d=None, **kwargs):
        if d is None:
            d = {}
        if kwargs:
            d.update(**kwargs)
        for k, v in d.items():
            setattr(self, k, v)

    def __setattr__(self, name, value):
        if name in _RESERVED_METHODS:
            raise ConfigError('"{}()" is a reserved method, cannot override'
                              .format(name))
        if isinstance(value, (list, tuple)):
            value = [self.__class__(x)
                     if isinstance(x, dict) else x for x in value]
        elif isinstance(value, dict) and not isinstance(value, self.__class__):
            value = self.__class__(value)
        super().__setattr__(name, value)
        super().__setitem__(name, value)

    __setitem__ = __setattr__

    def update(self, other):
        for k, v in other.items():
            setattr(self, k, v)

    def __getattr__(self, key):
        try:
            return super().__getattribute__(key)    
        except AttributeError:
            raise ConfigError('config key "{}" missing.'.format(key))

    @classmethod
    def load_file(cls, file_path):
        assert file_path.endswith('.json') or file_path.endswith('yaml')
        with open(file_path, 'r') as fp:
            if file_path.endswith('.json'):
                return cls(json.load(fp))
            else:
                return cls(yaml.load(fp))

    def to_dict(self):
        """
        Recursively convert back to dictionary
        """
        d = {}
        for k, value in self.items():
            if isinstance(value, Config):
                d[k] = value.to_dict()
            elif isinstance(value, (list, tuple)):
                d[k] = type(value)(
                    v.to_dict() if isinstance(v, Config) else v for v in value)
            else:
                d[k] = value
        return d

    def dump_file(self, file_path):
        assert file_path.endswith('.json') or file_path.endswith('yaml') or file_path.endswith('yml')
        with open(file_path, 'w') as fp:
            if file_path.endswith('.json'):
                json.dump(self, fp, indent=4)
            else:
                yaml.dump(
                    self.to_dict(),
                    stream=fp,
                    indent=4,
                    default_flow_style=False
                )

    def extend(self, default_config):
        U.assert_type(default_config, dict)
        return _fill_default_config(self, default_config, [])

    def copy(self):
        return Config(self.to_dict())


def extend_config(config, default_config):
    """
    default_config must specify all the expected keys. Use the following special
    values for required placeholders:

    * _req_: require a single value (not a list or dict)
    * _req_DICT_: require a dict
    * _req_LIST_: require a list

    Returns:
        AttributeDict
        `config` filled by default values if certain keys are unspecified

    Raises:
        ConfigError if required placeholders are not satisfied
    """
    U.assert_type(config, dict)
    U.assert_type(default_config, dict)
    return Config(_fill_default_config(config, default_config, []))



"""
File system utils.
"""
import os
import sys
import errno
import shutil
import glob
import pwd
import codecs
import hashlib
import tarfile
from socket import gethostname

f_ext = os.path.splitext

f_expand = os.path.expanduser

f_size = os.path.getsize

is_file = os.path.isfile

is_dir = os.path.isdir

get_dir = os.path.dirname


def owner_name(filepath):
    """
    Returns: file owner name, unix only
    """
    return pwd.getpwuid(os.stat(filepath).st_uid).pw_name


def host_name():
    "Get host name, alias with ``socket.gethostname()``"
    return gethostname()


def host_id():
    """
    Returns: first part of hostname up to '.'
    """
    return host_name().split('.')[0]


def utf_open(fname, mode):
    """
    Wrapper for codecs.open
    """
    return codecs.open(fname, mode=mode, encoding='utf-8')


def is_txt(fpath):
    "Test if file path is a text file"
    _, ext = f_ext(fpath)
    return ext == '.txt'


def f_exists(path):
    return os.path.exists(f_expand(path))


def f_join(*fpaths):
    """
    join file paths and expand special symbols like `~` for home dir
    """
    return f_expand(os.path.join(*fpaths))


def f_mkdir(fpath):
    """
    Recursively creates all the subdirs
    If exist, do nothing.
    """
    os.makedirs(f_expand(fpath), exist_ok=True)


def f_mkdir_in_path(fpath):
    """
    fpath is a file,
    recursively creates all the parent dirs that lead to the file
    If exist, do nothing.
    """
    os.makedirs(get_dir(f_expand(fpath)), exist_ok=True)


def f_last_part_in_path(fpath):
    """
    https://stackoverflow.com/questions/3925096/how-to-get-only-the-last-part-of-a-path-in-python
    """
    return os.path.basename(os.path.normpath(f_expand(fpath)))


def f_time(fpath):
    "File modification time"
    return str(os.path.getctime(fpath))


def f_append_before_ext(fpath, suffix):
    """
    Append a suffix to file name and retain its extension
    """
    name, ext = f_ext(fpath)
    return name + suffix + ext


def f_add_ext(fpath, ext):
    """
    Append an extension if not already there
    Args:
      ext: will add a preceding `.` if doesn't exist
    """
    if not ext.startswith('.'):
        ext = '.' + ext
    if fpath.endswith(ext):
        return fpath
    else:
        return fpath + ext


def f_remove(fpath):
    """
    If exist, remove. Supports both dir and file. Supports glob wildcard.
    """
    fpath = f_expand(fpath)
    for f in glob.glob(fpath):
        try:
            shutil.rmtree(f)
        except OSError as e:
            if e.errno == errno.ENOTDIR:
                try:
                    os.remove(f)
                except: # final resort safeguard
                    pass


def f_copy(fsrc, fdst):
    """
    If exist, remove. Supports both dir and file. Supports glob wildcard.
    """
    fsrc, fdst = f_expand(fsrc), f_expand(fdst)
    for f in glob.glob(fsrc):
        try:
            shutil.copytree(f, fdst)
        except OSError as e:
            if e.errno == errno.ENOTDIR:
                shutil.copy(f, fdst)


def f_move(fsrc, fdst):
    fsrc, fdst = f_expand(fsrc), f_expand(fdst)
    for f in glob.glob(fsrc):
        shutil.move(f, fdst)


def f_split_path(fpath, normpath=True):
    """
    Splits path into a list of its component folders

    Args:
        normpath: call os.path.normpath to remove redundant '/' and
            up-level references like ".."
    """
    if normpath:
        fpath = os.path.normpath(fpath)
    allparts = []
    while 1:
        parts = os.path.split(fpath)
        if parts[0] == fpath:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == fpath:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            fpath = parts[0]
            allparts.insert(0, parts[1])
    return allparts


def script_dir():
    """
    Returns: the dir of current script
    """
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def parent_dir(location, abspath=False):
    """
    Args:
      location: current directory or file

    Returns:
        parent directory absolute or relative path
    """
    _path = os.path.abspath if abspath else os.path.relpath
    return _path(f_join(location, os.pardir))


def f_md5(fpath):
    """
    File md5 signature
    """
    hash_md5 = hashlib.md5()
    with open(f_expand(fpath), "rb") as f:
        for chunk in iter(lambda: f.read(65536), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def compress_tar(source_file, output_tarball, compress_mode='gz'):
    """
    Args:
        source_file: source file or folder
        output_tarball: output tar file name
        compress_mode: "gz", "bz2", "xz" or "" (empty for uncompressed write)
    """
    source_file, output_tarball = f_expand(source_file), f_expand(output_tarball)
    assert compress_mode in ['gz', 'bz2', 'xz', '']
    with tarfile.open(output_tarball, 'w:'+compress_mode) as tar:
        tar.add(source_file, arcname=os.path.basename(source_file))


def extract_tar(source_tarball, output_dir='.', members=None):
    """
    Args:
        source_tarball: extract members from archive
        output_dir: default to current working dir
        members: must be a subset of the list returned by getmembers()
    """
    source_tarball, output_dir = f_expand(source_tarball), f_expand(output_dir)
    with tarfile.open(source_tarball, 'r:*')  as tar:
        tar.extractall(output_dir, members=members)


def move_with_backup(path, suffix='.bak'):
    """
    Ensures that a path is not occupied. If there is a file, rename it by
    adding @suffix. Resursively backs up everything.

    Args:
        path: file path to clear
        suffix: Add to backed up files (default: {'.bak'})
    """
    path = str(path)
    if os.path.exists(path):
        move_with_backup(path + suffix)
        shutil.move(path, path + suffix)



import torch
import torch.nn as nn
import torchx as tx
from torchx.nn.hyper_scheduler import *
import numpy as np
from .base import Learner
from .aggregator import MultistepAggregatorWithInfo 
from surreal.model.ppo_net import PPOModel, DiagGauss
from surreal.model.reward_filter import RewardFilter
from surreal.session import Config, extend_config, BASE_SESSION_CONFIG, BASE_LEARNER_CONFIG, ConfigError

class PPOLearner(Learner):
    '''
    PPOLearner: subclass of Learner that contains PPO algorithm logic
    Attributes:
        gpu_option: 'cpu' if not using GPU, 'cuda:all' otherwise
        model: instance of PPOModel from surreal.model.ppo_net
        ref_target_model: instance of PPOModel, kept to used as
            reference policy
        ppo_mode: string of either 'adapt' or 'clip' to determine
            which variant of PPO is used. For details of variants
            see https://arxiv.org/pdf/1707.06347.pdf
        norm_adv: boolean flag -- whether to use batch advantage
            normalization
        use_z_filter: boolean flag -- whether to use obs Z-Filtering
        actor/critic_optim: Adam Optimizer for policy and baseline network
        actor/critic_lr_scheduler: Learning rate scheduler. details see
            surreal.utils.pytorch.scheduler
        aggregator: experience aggregator used to batch experiences.
            for available aggregators, see surreal.learner.aggregator
        pd: probability distribution class (Assumed as Diagonal Gaussian)
            see surreal.model.ppo_net for details

    important member functions:
        private methods:
        _clip_loss: computes the loss and various statistics
            for 'clip' variant PPO
        _clip_update: uses loss information to make policy update
        _adapt_loss: computes loss and various statistics for
            'adapt' variant of PPO
        _adapt_update: uses loss info to make policy update
        _value_loss: computes loss and various statistics for value function
        _value_update: uses loss info to update value function
        _gae_and_return: computes generalized advantage estimate and
            corresponding N-step return. Details of algorithm can be found
            here: https://arxiv.org/pdf/1506.02438.pdf
        _advantage_and_return: basic advantage and N-step return estimate
        _optimize: fucntion that makes policy and value function update
        _post_publish: function that manages metrics and behavior after
            parameter release

        public methods:
        learn: method to perform optimization and send to tensorplex for log
        module_dict: returns the corresponding parameters
        publish_parameter: publishes parameters in self.model to parameter server
    '''
    def __init__(self, learner_config, env_config, session_config):
        super().__init__(learner_config, env_config, session_config)

        # GPU setting
        self.current_iteration = 0
        self.global_step = 0
        if not torch.cuda.is_available():
            self.gpu_option = 'cpu'
        else:
            self.gpu_option = 'cuda:all'
        self.use_cuda = torch.cuda.is_available()

        if not self.use_cuda:
            self.log.info('Using CPU')
        else:
            self.log.info('Using GPU: {}'.format(self.gpu_option)) 

        # RL general parameters
        self.gamma = self.learner_config.algo.gamma
        self.lam   = self.learner_config.algo.advantage.lam
        self.n_step = self.learner_config.algo.n_step
        self.use_z_filter = self.learner_config.algo.use_z_filter
        self.use_r_filter = self.learner_config.algo.use_r_filter
        self.norm_adv = self.learner_config.algo.advantage.norm_adv
        self.batch_size = self.learner_config.replay.batch_size

        self.action_dim = self.env_config.action_spec.dim[0]
        self.obs_spec = self.env_config.obs_spec                                           
        self.init_log_sig = self.learner_config.algo.consts.init_log_sig

        # PPO parameters
        self.ppo_mode = self.learner_config.algo.ppo_mode
        self.if_rnn_policy = self.learner_config.algo.rnn.if_rnn_policy
        self.horizon = self.learner_config.algo.rnn.horizon
        self.lr_actor = self.learner_config.algo.network.lr_actor
        self.lr_critic = self.learner_config.algo.network.lr_critic
        self.epoch_policy = self.learner_config.algo.consts.epoch_policy
        self.epoch_baseline = self.learner_config.algo.consts.epoch_baseline
        self.kl_target = self.learner_config.algo.consts.kl_target
        self.adjust_threshold = self.learner_config.algo.consts.adjust_threshold
        self.reward_scale = self.learner_config.algo.advantage.reward_scale

        # PPO mode 'adjust'
        self.kl_cutoff_coeff = self.learner_config.algo.adapt_consts.kl_cutoff_coeff
        self.beta_init = self.learner_config.algo.adapt_consts.beta_init
        self.beta_range = self.learner_config.algo.adapt_consts.beta_range

        # PPO mode 'clip'
        self.clip_range = self.learner_config.algo.clip_consts.clip_range
        self.clip_epsilon_init = self.learner_config.algo.clip_consts.clip_epsilon_init

        if self.ppo_mode == 'adapt':
            self.beta = self.beta_init
            self.eta = self.kl_cutoff_coeff
            self.beta_upper = self.beta_range[1]
            self.beta_lower = self.beta_range[0]
            self.beta_adjust_threshold = self.adjust_threshold
        else: # method == 'clip'
            self.clip_epsilon = self.clip_epsilon_init
            self.clip_adjust_threshold = self.adjust_threshold
            self.clip_upper = self.clip_range[1]
            self.clip_lower = self.clip_range[0]

        # learning rate setting:
        self.min_lr = self.learner_config.algo.network.anneal.min_lr
        self.lr_update_frequency = self.learner_config.algo.network.anneal.lr_update_frequency
        self.frames_to_anneal = self.learner_config.algo.network.anneal.frames_to_anneal
        num_updates = int(self.frames_to_anneal / self.learner_config.parameter_publish.exp_interval)
        lr_scheduler = eval(self.learner_config.algo.network.anneal.lr_scheduler) 

        self.exp_counter = 0
        self.kl_record = []

        with tx.device_scope(self.gpu_option):
            self.model = PPOModel(
                obs_spec=self.obs_spec,
                action_dim=self.action_dim,
                model_config=self.learner_config.model,
                use_cuda=self.use_cuda,
                init_log_sig=self.init_log_sig,
                use_z_filter=self.use_z_filter,
                if_pixel_input=self.env_config.pixel_input,
                rnn_config=self.learner_config.algo.rnn,
            )
            self.ref_target_model = PPOModel(
                obs_spec=self.obs_spec,
                action_dim=self.action_dim,
                model_config=self.learner_config.model,
                use_cuda=self.use_cuda,
                init_log_sig=self.init_log_sig,
                use_z_filter=self.use_z_filter,
                if_pixel_input=self.env_config.pixel_input,
                rnn_config=self.learner_config.algo.rnn,
            )
            self.ref_target_model.update_target_params(self.model)

            # Learning parameters and optimizer
            self.clip_actor_gradient = self.learner_config.algo.network.clip_actor_gradient
            self.actor_gradient_clip_value = self.learner_config.algo.network.actor_gradient_norm_clip
            self.clip_critic_gradient = self.learner_config.algo.network.clip_critic_gradient
            self.critic_gradient_clip_value = self.learner_config.algo.network.critic_gradient_norm_clip

            self.critic_optim = torch.optim.Adam(
                self.model.get_critic_params(),
                lr=self.lr_critic,
                weight_decay=self.learner_config.algo.network.critic_regularization
            )
            self.actor_optim = torch.optim.Adam(
                self.model.get_actor_params(),
                lr=self.lr_actor,
                weight_decay=self.learner_config.algo.network.actor_regularization
            )

            # learning rate scheduler
            self.actor_lr_scheduler  = lr_scheduler(self.actor_optim, 
                                                    num_updates,
                                                    update_freq=self.lr_update_frequency,
                                                    min_lr = self.min_lr)
            self.critic_lr_scheduler = lr_scheduler(self.critic_optim, 
                                                    num_updates,
                                                    update_freq=self.lr_update_frequency,
                                                    min_lr = self.min_lr)

            # Experience Aggregator
            self.aggregator = MultistepAggregatorWithInfo(self.env_config.obs_spec, 
                                                          self.env_config.action_spec)
        
            # probability distribution. Gaussian only for now
            self.pd = DiagGauss(self.action_dim)

            # placeholder for RNN hidden cells
            self.cells = None

            # Reward White-filtering
            if self.use_r_filter: 
                self.reward_filter= RewardFilter()

    def _clip_loss(self, obs, actions, advantages, behave_pol): 
        """
        Computes the loss with current data. also returns a dictionary of statistics
        which includes surrogate loss, clipped surrogate los, policy entropy, clip
        constant
        return: surreal.utils.pytorch.GPUVariable, dict
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            actions: batch of actions in form of (batch_size, act_dim)
            advantages: batch of normalized advantage, (batch_size, 1)
            behave_pol: batch of behavior policy (batch_size, 2 * act_dim)
        Returns:
            clip_loss: Variable for loss
            stats: dictionary of recorded statistics
        """
        learn_pol = self.model.forward_actor(obs, self.cells)
        learn_prob = self.pd.likelihood(actions, learn_pol)
        behave_prob = self.pd.likelihood(actions, behave_pol)
        prob_ratio = learn_prob / behave_prob
        cliped_ratio = torch.clamp(prob_ratio, 1 - self.clip_epsilon, 
                                               1 + self.clip_epsilon)
        surr = -prob_ratio * advantages.view(-1, 1)
        cliped_surr = -cliped_ratio * advantages.view(-1, 1)
        clip_loss = torch.cat([surr, cliped_surr], 1).max(1)[0].mean() 

        stats = {
            "_surr_loss": surr.mean().item(), 
            "_clip_surr_loss": clip_loss.item(), 
            "_entropy": self.pd.entropy(learn_pol).mean().item(),
            '_clip_epsilon': self.clip_epsilon
        }
        return clip_loss, stats

    def _clip_update(self, obs, actions, advantages, behave_pol):
        """
        Method that makes policy updates. calls _clip_loss method
        Note:  self.clip_actor_gradient determines whether gradient is clipped
        return: dictionary of statistics to be sent to tensorplex server
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            actions: batch of actions in form of (batch_size, act_dim)
            advantages: batch of normalized advantage, (batch_size, 1)
            behave_pol: batch of behavior policy (batch_size, 2 * act_dim)
        Returns:
            stats: dictionary of recorded statistics
        """
        loss, stats = self._clip_loss(obs, actions, advantages, behave_pol)
        self.model.clear_actor_grad()
        loss.backward()
        if self.clip_actor_gradient:
            stats['grad_norm_actor'] = nn.utils.clip_grad_norm_(
                                            self.model.get_actor_params(), 
                                            self.actor_gradient_clip_value).item()
        self.actor_optim.step()
        return stats

    def _adapt_loss(self, obs, actions, advantages, behave_pol, ref_pol):
        """
        Computes the loss with current data. also returns a dictionary of statistics
        which includes surrogate loss, clipped surrogate los, policy entropy, adaptive
        KL penalty constant, policy KL divergence
        return: surreal.utils.pytorch.GPUVariable, dict
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            actions: batch of actions in form of (batch_size, act_dim)
            advantages: batch of normalized advantage, (batch_size, 1)
            behave_pol: batch of behavior policy (batch_size, 2 * act_dim)
            ref_pol: batch of reference policy (batch_size, 2 * act_dim)
        Returns:
            loss: Variable for loss
            stats: dictionary of recorded statistics
        """
        learn_pol = self.model.forward_actor(obs, self.cells)
        prob_behave = self.pd.likelihood(actions, behave_pol)
        prob_learn  = self.pd.likelihood(actions, learn_pol)
        
        kl = self.pd.kl(ref_pol, learn_pol).mean()
        surr = -(advantages.view(-1, 1) * (prob_learn/ torch.clamp(prob_behave, min=1e-2))).mean()
        loss = surr + self.beta * kl
        entropy = self.pd.entropy(learn_pol).mean()

        if kl.item() - 2.0 * self.kl_target > 0:
            loss += self.eta * (kl - 2.0 * self.kl_target).pow(2)

        stats = {
            '_kl_loss_adapt': loss.item(), 
            '_surr_loss': surr.item(), 
            '_pol_kl': kl.item(), 
            '_entropy': entropy.item(),
            '_beta': self.beta
        }
        return loss, stats

    def _adapt_update(self, obs, actions, advantages, behave_pol, ref_pol):
        """
        Method that makes policy updates. calls _adapt_loss method
        Note:  self.clip_actor_gradient determines whether gradient is clipped
        return: dictionary of statistics to be sent to tensorplex server
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            actions: batch of actions in form of (batch_size, act_dim)
            advantages: batch of normalized advantage, (batch_size, 1)
            behave_pol: batch of behavior policy (batch_size, 2 * act_dim)
            ref_pol: batch of reference policy (batch_size, 2 * act_dim)
        Returns:
            stats: dictionary of recorded statistics
        """
        loss, stats = self._adapt_loss(obs, actions, advantages, behave_pol, ref_pol)
        self.model.clear_actor_grad()
        loss.backward()
        if self.clip_actor_gradient:
            stats['grad_norm_actor'] = nn.utils.clip_grad_norm_(
                                            self.model.get_actor_params(), 
                                            self.actor_gradient_clip_value).item()
        self.actor_optim.step()
        return stats

    def _value_loss(self, obs, returns):
        """
        Computes the loss with current data. also returns a dictionary of statistics
        which includes value loss and explained variance
        return: surreal.utils.pytorch.GPUVariable, dict
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            returns: batch of N-step return estimate (batch_size,)
        Returns:
            loss: Variable for loss
            stats: dictionary of recorded statistics
        """
        values = self.model.forward_critic(obs, self.cells)
        if len(values.size()) == 3: values = values.squeeze(2)
        explained_var = 1 - torch.var(returns - values) / torch.var(returns)
        loss = (values - returns).pow(2).mean()

        stats = {
            '_val_loss': loss.item(),
            '_val_explained_var': explained_var.item()
        }
        return loss, stats

    def _value_update(self, obs, returns):
        """
        Method that makes baseline function updates. calls _value_loss method
        Note:  self.clip_actor_gradient determines whether gradient is clipped
        return: dictionary of statistics to be sent to tensorplex server
        Args:
            obs: batch of observations in form of (batch_size, obs_dim)
            returns: batch of N-step return estimate (batch_size,)
        Returns:
            stats: dictionary of recorded statistics
        """
        loss, stats = self._value_loss(obs, returns)
        self.model.clear_critic_grad()
        loss.backward()
        if self.clip_critic_gradient:
            stats['grad_norm_critic'] = nn.utils.clip_grad_norm_(
                                                self.model.get_critic_params(), 
                                                self.critic_gradient_clip_value).item()
        self.critic_optim.step()
        return stats

    def _gae_and_return(self, obs, obs_next, rewards, dones):
        '''
        computes generalized advantage estimate and corresponding N-step return. 
        Details of algorithm can be found here: https://arxiv.org/pdf/1506.02438.pdf
        Args: 
            obs: batch of observations (batch_size, N-step , obs_dim)
            obs_next: batch of next observations (batch_size, 1 , obs_dim)
            actions: batch of actions (batch_size, N-step , act_dim)
            rewards: batch of rewards (batch_size, N-step)
            dones: batch of termination flags (batch_size, N-step)
        Returns:
            obs: batch of observation (batch_size, obs_dim)
            actions: batch of action (batch_size, act_dim)
            advantage: batch of advantages (batch_size, 1)
            returns: batch of returns (batch_size, 1)
        '''
        with tx.device_scope(self.gpu_option):
            index_set = torch.tensor(range(self.n_step), dtype=torch.float32)
            gamma = torch.pow(self.gamma, index_set)
            lam = torch.pow(self.lam, index_set)

            obs_concat_var = {}
            for mod in obs.keys():
                obs_concat_var[mod] = {}
                for k in obs[mod].keys():
                    obs_concat_var[mod][k] = (torch.cat([obs[mod][k], obs_next[mod][k]], dim=1))
                    if not self.if_rnn_policy:
                        obs_shape = obs_concat_var[mod][k].size()
                        obs_concat_var[mod][k] = obs_concat_var[mod][k].view(-1, *obs_shape[2:])

            values = self.model.forward_critic(obs_concat_var, self.cells) 
            values = values.view(self.batch_size, self.n_step + 1)
            values[:, 1:] *= 1 - dones

            if self.if_rnn_policy:
                tds = rewards + self.gamma * values[:, 1:] - values[:, :-1]
                eff_len = self.n_step - self.horizon + 1
                gamma = gamma[:self.horizon]
                lam = lam[:self.horizon]

                returns = torch.zeros(self.batch_size, eff_len)
                advs = torch.zeros(self.batch_size, eff_len)
                for step in range(eff_len):
                    returns[:, step] = torch.sum(gamma * rewards[:, step:step + self.horizon], 1) + \
                                       values[:, step + self.horizon] * (self.gamma ** self.horizon)
                    advs[:, step] = torch.sum(tds[:, step:step + self.horizon] * gamma * lam, 1)

                if self.norm_adv:
                    std = advs.std()
                    mean = advs.mean()
                    advs = (advs - mean) / max(std, 1e-4)
                return advs, returns

            else:
                returns = torch.sum(gamma * rewards, 1) + values[:, -1] * (self.gamma ** self.n_step)
                tds = rewards + self.gamma * values[:, 1:] - values[:, :-1] 
                gae = torch.sum(tds * gamma * lam, 1)

                if self.norm_adv:
                    std = gae.std()
                    mean = gae.mean()
                    gae = (gae - mean) / max(std, 1e-4)

                return gae.view(-1, 1), returns.view(-1, 1)

    def _preprocess_batch_ppo(self, batch):
        '''
            Loading experiences from numpy to torch.FloatTensor type
            Args: 
                batch: BeneDict of experiences containing following attributes
                        'obs' - observation
                        'actions' - actions
                        'rewards' - rewards
                        'obs_next' - next observation
                        'persistent_infos' - action policy
                        'onetime_infos' - RNN hidden cells or None
            Return:
                Benedict of torch.FloatTensors
        '''
        with tx.device_scope(self.gpu_option):

            obs, actions, rewards, obs_next, done, persistent_infos, onetime_infos = (
                batch['obs'],
                batch['actions'],
                batch['rewards'],
                batch['obs_next'],
                batch['dones'],
                batch['persistent_infos'],
                batch['onetime_infos'],
            )

            for modality in obs:
                for key in obs[modality]:
                    obs[modality][key] = (torch.tensor(obs[modality][key], dtype=torch.float32)).detach()
                    obs_next[modality][key] = (torch.tensor(obs_next[modality][key], dtype=torch.float32)).detach()

            actions = torch.tensor(actions, dtype=torch.float32)
            rewards = torch.tensor(rewards, dtype=torch.float32) * self.reward_scale
            if self.use_r_filter:
                normed_reward = self.reward_filter.forward(rewards)
                self.reward_filter.update(rewards)
                rewards = normed_reward

            done = torch.tensor(done, dtype=torch.float32)

            if persistent_infos is not None:
                for i in range(len(persistent_infos)):
                    persistent_infos[i] = torch.tensor(persistent_infos[i], dtype=torch.float32).detach()
            if onetime_infos is not None:
                for i in range(len(onetime_infos)):
                    onetime_infos[i] = torch.tensor(onetime_infos[i], dtype=torch.float32).detach()

            (
                batch['obs'],
                batch['actions'],
                batch['rewards'],
                batch['obs_next'],
                batch['dones'],
                batch['persistent_infos'],
                batch['onetime_infos'],
            ) = (
                obs,
                actions,
                rewards,
                obs_next,
                done,
                persistent_infos,
                onetime_infos
            )
            return batch


    def _optimize(self, obs, actions, rewards, obs_next, persistent_infos, onetime_infos, dones):
        '''
            main method for optimization that calls _adapt/clip_update and 
            _value_update epoch_policy and epoch_baseline times respectively
            return: dictionary of tracted statistics
            Args:
                obs: batch of observations (batch_size, N-step , obs_dim)
                obs_next: batch of next observations (batch_size, 1 , obs_dim)
                actions: batch of actions (batch_size, N-step , act_dim)
                rewards: batch of rewards (batch_size, N-step)
                dones: batch of termination flags (batch_size, N-step)
                action_infos: list of batched other attributes tracted, such as
                    behavior policy, RNN hidden states and etc.
            Returns:
                dictionary of recorded statistics
        '''
        # convert everything to float tensor: 
        with tx.device_scope(self.gpu_option):
            pds = persistent_infos[-1]

            if self.if_rnn_policy:
                h = (onetime_infos[0].transpose(0, 1).contiguous()).detach()
                c = (onetime_infos[1].transpose(0, 1).contiguous()).detach()
                self.cells = (h, c)

            advantages, returns = self._gae_and_return(obs, 
                                                       obs_next,  
                                                       rewards, 
                                                       dones)
            advantages = advantages.detach()
            returns    = returns.detach()

            if self.if_rnn_policy:
                h = self.cells[0].detach()
                c = self.cells[1].detach()
                self.cells = (h, c)
                eff_len = self.n_step - self.horizon + 1
                behave_pol = pds[:, :eff_len, :].contiguous().detach()
                actions_iter = actions[:, :eff_len, :].contiguous().detach()
            else:
                behave_pol = pds[:, 0, :].contiguous().detach()
                actions_iter = actions[:, 0, :].contiguous().detach()

            obs_iter = {}
            for mod in obs.keys():
                obs_iter[mod] = {}
                for k in obs[mod].keys():
                    if self.if_rnn_policy:
                        obs_iter[mod][k] = obs[mod][k][:, :self.n_step - self.horizon + 1, :].contiguous().detach()
                    else: 
                        obs_iter[mod][k] = obs[mod][k][:, 0, :].contiguous().detach()

            ref_pol = self.ref_target_model.forward_actor(obs_iter, self.cells).detach()

            for ep in range(self.epoch_policy):
                if self.ppo_mode == 'clip':
                    stats =  self._clip_update(obs_iter, 
                                               actions_iter, 
                                               advantages, 
                                               behave_pol)
                else: 
                    stats = self._adapt_update(obs_iter, 
                                               actions_iter, 
                                               advantages, 
                                               behave_pol, 
                                               ref_pol)
                curr_pol = self.model.forward_actor(obs_iter, self.cells).detach()
                kl = self.pd.kl(ref_pol, curr_pol).mean()
                stats['_pol_kl'] = kl.item()
                if kl.item() > self.kl_target * 4: 
                    break

            self.kl_record.append(stats['_pol_kl'])

            for _ in range(self.epoch_baseline):
                baseline_stats = self._value_update(obs_iter, returns)

            # Collecting metrics and updating tensorplex
            for k in baseline_stats:
                stats[k] = baseline_stats[k]

            behave_likelihood = self.pd.likelihood(actions_iter, behave_pol)
            curr_likelihood   = self.pd.likelihood(actions_iter, curr_pol)

            stats['_avg_return_targ'] = returns.mean().item()
            stats['_avg_log_sig'] = self.model.actor.log_var.mean().item()
            stats['_avg_behave_likelihood'] = behave_likelihood.mean().item()
            stats['_avg_is_weight'] = (curr_likelihood / (behave_likelihood + 1e-4)).mean().item()
            stats['_ref_behave_diff'] = self.pd.kl(ref_pol, behave_pol).mean().item()
            stats['_lr'] = self.actor_lr_scheduler.get_lr()[0]

            if self.use_z_filter:
                self.model.z_update(obs_iter)
                stats['obs_running_mean'] = np.mean(self.model.z_filter.running_mean())
                stats['obs_running_square'] =  np.mean(self.model.z_filter.running_square())
                stats['obs_running_std'] = np.mean(self.model.z_filter.running_std())
            if self.use_r_filter:
                stats['reward_mean'] = self.reward_filter.reward_mean()

            return stats

    def learn(self, batch):
        '''
            main method for learning, calls _optimize. Also sends update stats 
            to Tensorplex
            Args:
                batch: pre-aggregated list of experiences rolled out by the agent
        '''
        self.current_iteration += 1
        batch = self._preprocess_batch_ppo(batch)
        tensorplex_update_dict = self._optimize(
            batch.obs,
            batch.actions,
            batch.rewards,
            batch.obs_next,
            batch.persistent_infos,
            batch.onetime_infos,
            batch.dones,
        )
        self.periodic_checkpoint(
            global_steps=self.current_iteration,
            score=None,
        )

        self.tensorplex.add_scalars(tensorplex_update_dict, self.global_step)
        self.exp_counter += self.batch_size
        self.global_step += 1

    def module_dict(self):
        '''
        returns the corresponding parameters
        '''
        return {
            'ppo': self.model,
        }

    def publish_parameter(self, iteration, message=''):
        """
        Learner publishes latest parameters to the parameter server only when 
        accumulated enough experiences specified by 
            learner_config.algo.network.update_target.interval
        Note: this overrides the base class publish_parameter method
        Args:
            iteration: the current number of learning iterations
            message: optional message, must be pickleable.
        """
        if  self.exp_counter >= self.learner_config.parameter_publish.exp_interval:
            self._ps_publisher.publish(iteration, message=message)
            self._post_publish()  

    def _post_publish(self):
        '''
            function that manages metrics and behavior after parameter release
            Actions include: 
                adjusts adaptive threshold for KL penalty for 'adapt' PPO 
                adjusts adaptive prob ratio clip rate for 'clip' PPO
                clears KL-Divergence record
                clears experience counter after parameter release
                steps actor and critic learning rate scheduler
        '''
        final_kl = np.mean(self.kl_record)
        if self.ppo_mode == 'clip': # adapts clip ratios
            if final_kl > self.kl_target * self.clip_adjust_threshold[1]:
                if self.clip_lower < self.clip_epsilon:
                    self.clip_epsilon = self.clip_epsilon / self.learner_config.algo.clip_consts.scale_constant
            elif final_kl < self.kl_target * self.clip_adjust_threshold[0]:
                if self.clip_upper > self.clip_epsilon:
                    self.clip_epsilon = self.clip_epsilon * self.learner_config.algo.clip_consts.scale_constant
        else: # adapt KL divergence penalty before returning the statistics 
            if final_kl > self.kl_target * self.beta_adjust_threshold[1]:
                if self.beta_upper > self.beta:
                    self.beta = self.beta * self.learner_config.algo.adapt_consts.scale_constant
            elif final_kl < self.kl_target * self.beta_adjust_threshold[0]:
                if self.beta_lower < self.beta:
                    self.beta = self.beta / self.learner_config.algo.adapt_consts.scale_constant
        self.ref_target_model.update_target_params(self.model)
        self.kl_record = []
        self.exp_counter = 0
        self.actor_lr_scheduler.step()
        self.critic_lr_scheduler.step()

    def checkpoint_attributes(self):
        '''
            outlines attributes to be checkpointed
        '''
        return [
            'model',
            'ref_target_model',
            'actor_lr_scheduler',
            'critic_lr_scheduler',
            'current_iteration',
        ]

    def _prefetcher_preprocess(self, batch):
        batch = self.aggregator.aggregate(batch)
        return batch



import torch.nn as nn
from torch.nn.init import xavier_uniform
import surreal.utils as U
import torch.nn.functional as F
import numpy as np
import torchx.nn as nnx

from .model_builders import *
from .z_filter import ZFilter

import itertools

class DiagGauss(object):
    '''
        Class that encapsulates Diagonal Gaussian Probability distribution
        Attributes:
            d: action dimension
        Member Functions:
            loglikelihood
            likelihood
            kl
            entropy
            sample
            maxprob
    '''
    def __init__(self, action_dim):
        self.d = action_dim

    def loglikelihood(self, a, prob):
        '''
            Method computes loglikelihood of action (a) given probability (prob)
        '''
        if len(a.size()) == 3:
            a = a.view(-1, self.d)
            prob = prob.view(-1, 2 * self.d)

        mean0 = prob[:, :self.d]
        std0 = prob[:, self.d:]
        return - 0.5 * (((a - mean0) / std0).pow(2)).sum(dim=1, keepdim=True) - 0.5 * np.log(
            2.0 * np.pi) * self.d - std0.log().sum(dim=1, keepdim=True)

    def likelihood(self, a, prob):
        '''
            Method computes likelihood of action (a) given probability (prob)
        '''
        return torch.clamp(self.loglikelihood(a, prob).exp(), min=1e-5)

    def kl(self, prob0, prob1):
        '''
            Method computes KL Divergence of between two probability distributions
            Note: this is D_KL(prob0 || prob1), not D_KL(prob1 || prob0)
        '''
        if len(prob0.size()) == 3:
            prob0 = prob0.view(-1, 2 * self.d)
            prob1 = prob1.view(-1, 2 * self.d)

        mean0 = prob0[:, :self.d]
        std0 = prob0[:, self.d:]
        mean1 = prob1[:, :self.d]
        std1 = prob1[:, self.d:]
        return ((std1 / std0).log()).sum(dim=1) + (
            (std0.pow(2) + (mean0 - mean1).pow(2)) / (2.0 * std1.pow(2))).sum(dim=1) - 0.5 * self.d

    def entropy(self, prob):
        '''
            Method computes entropy of a given probability (prob)
        '''
        if len(prob.size()) == 3:
            prob = prob.view(-1, 2 * self.d)

        std_nd = prob[:, self.d:]
        return 0.5 * std_nd.log().sum(dim=1) + .5 * np.log(2 * np.pi * np.e) * self.d

    def sample(self, prob):
        '''
            Method samples actions from probability distribution
        '''
        if len(prob.shape) == 3:
            prob_shape = prob.shape
            prob = prob.reshape(-1, self.d * 2)
        mean_nd = prob[:, :self.d]
        std_nd = prob[:, self.d:]
        return np.random.randn(prob.shape[0], self.d) * std_nd + mean_nd

    def maxprob(self, prob):
        '''
            Method deterministically sample actions of maximum likelihood
        '''
        if len(prob.shape) == 3:
            return prob[:, :, self.d]
        return prob[:, :self.d]


class PPOModel(nnx.Module):
    '''
        PPO Model class that wraps aroud the actor and critic networks
        Attributes:
            actor: Actor network, see surreal.model.model_builders.builders
            critic: Critic network, see surreal.model.model_builders.builders
            z_filter: observation z_filter. see surreal.model.z_filter
        Member functions:
            update_target_param: updates kept parameters to that of another model
            update_target_param: updates kept z_filter to that of another model
            forward_actor: forward pass actor to generate policy with option
                to use z-filter
            forward_actor: forward pass critic to generate policy with option
                to use z-filter
            z_update: updates Z_filter running obs mean and variance
    '''
    def __init__(self,
                 obs_spec,
                 action_dim,
                 model_config,
                 use_cuda,
                 init_log_sig=0,
                 use_z_filter=False,
                 if_pixel_input=False,
                 rnn_config=None):
        super(PPOModel, self).__init__()

        # hyperparameters
        self.obs_spec = obs_spec
        self.action_dim = action_dim
        self.model_config = model_config
        self.use_z_filter = use_z_filter
        self.init_log_sig = init_log_sig
        self.if_pixel_input = if_pixel_input
        self.rnn_config = rnn_config

        # compute low dimensional feature dimension
        self.low_dim = 0
        if 'low_dim' in self.obs_spec.keys():
            for key in self.obs_spec['low_dim'].keys():
                self.low_dim += self.obs_spec['low_dim'][key][0]

        # optional CNN stem feature extractor
        self.cnn_stem = None
        if self.if_pixel_input:
            self.cnn_stem = CNNStemNetwork(self.obs_spec['pixel']['camera0'],
                                           self.model_config.cnn_feature_dim)

        # optional LSTM stem feature extractor
        self.rnn_stem = None
        if self.rnn_config.if_rnn_policy:
            rnn_insize = self.low_dim + (self.model_config.cnn_feature_dim if self.if_pixel_input else 0)
            self.rnn_stem = nn.LSTM(rnn_insize,
                                    self.rnn_config.rnn_hidden,
                                    self.rnn_config.rnn_layer,
                                    batch_first=True)
            if use_cuda:
                device = torch.device("cuda")
                self.rnn_stem = self.rnn_stem.to(device)

        # computing final feature dimension for leaf actor/critic network
        input_size = self.low_dim + (self.model_config.cnn_feature_dim if self.if_pixel_input else 0)
        input_size = self.rnn_config.rnn_hidden if self.rnn_config.if_rnn_policy else input_size

        self.actor = PPO_ActorNetwork(input_size, 
                                      self.action_dim, 
                                      self.model_config.actor_fc_hidden_sizes,
                                      self.init_log_sig)
        self.critic = PPO_CriticNetwork(input_size,
                                      self.model_config.critic_fc_hidden_sizes)
        if self.use_z_filter:
            assert self.low_dim > 0, "No low dimensional input, please turn off z-filter"
            self.z_filter = ZFilter(self.obs_spec)

    def _gather_low_dim_input(self, obs):
        '''
            Concatenate (along 2nd dimension) all the low-dimensional
            (propioceptive) features from input observation tuple
            Args:
                obs: dictionary of tensors
        '''
        if 'low_dim' not in obs.keys(): return None
        list_obs_ld = [obs['low_dim'][key] for key in obs['low_dim'].keys()]
        obs_low_dim = torch.cat(list_obs_ld, -1)
        return obs_low_dim

    def clear_actor_grad(self):
        '''
            Method that clears all gradients from all the parameters from
            actor, cnn_stem (optional), and rnn_stem (optional)
        '''
        self.actor.zero_grad()
        if self.if_pixel_input:
            self.cnn_stem.zero_grad()
        if self.rnn_config.if_rnn_policy:
            self.rnn_stem.zero_grad()

    def clear_critic_grad(self):
        '''
            Method that clears all gradients from all the parameters from
            critic, cnn_stem (optional), and rnn_stem (optional)
        '''
        self.critic.zero_grad()
        if self.if_pixel_input:
            self.cnn_stem.zero_grad()
        if self.rnn_config.if_rnn_policy:
            self.rnn_stem.zero_grad()

    def get_actor_params(self):
        '''
            Method that returns generator that contains all the parameters from
            actor, cnn_stem (optional), and rnn_stem (optional)
        '''
        params = self.actor.parameters()
        if self.if_pixel_input:
            params = itertools.chain(params, self.cnn_stem.parameters())
        if self.rnn_config.if_rnn_policy:
            params = itertools.chain(params, self.rnn_stem.parameters())
        return params

    def get_critic_params(self):
        '''
            Method that returns generator that contains all the parameters from
            critic, cnn_stem (optional), and rnn_stem (optional)
        '''
        params = self.critic.parameters()
        if self.if_pixel_input:
            params = itertools.chain(params, self.cnn_stem.parameters())
        if self.rnn_config.if_rnn_policy:
            params = itertools.chain(params, self.rnn_stem.parameters())
        return params

    def update_target_params(self, net):
        '''
            updates kept parameters to that of another model
            Args:
                net: another PPO_Model instance
        '''
        self.actor.load_state_dict(net.actor.state_dict())
        self.critic.load_state_dict(net.critic.state_dict())

        if self.rnn_config.if_rnn_policy:
            self.rnn_stem.load_state_dict(net.rnn_stem.state_dict())

        if self.if_pixel_input:
            self.cnn_stem.load_state_dict(net.cnn_stem.state_dict())

        if self.use_z_filter:
            self.z_filter.load_state_dict(net.z_filter.state_dict())

    def update_target_z_filter(self, net):
        '''
            updates kept z-filter to that of another model
            Args:
                net: another PPO_Model instance
        '''
        if self.use_z_filter:
            self.z_filter.load_state_dict(net.z_filter.state_dict())

    def forward_actor(self, obs, cells=None):
        '''
            forward pass actor to generate policy with option to use z-filter
            Args:
                obs: dictionary of batched observations
                cells: tuple of hidden and cell state for LSTM stem. optional
            Returns:
                The output of actor network
        '''
        obs_list = []
        obs_flat = self._gather_low_dim_input(obs)
        if self.use_z_filter:
            obs_flat = self.z_filter.forward(obs_flat)
        obs_list.append(obs_flat)

        if self.if_pixel_input:
            # right now assumes only one camera angle.
            obs_pixel = obs['pixel']['camera0']
            obs_pixel = self._scale_image(obs_pixel)
            obs_pixel = self.cnn_stem(obs_pixel)
            obs_list.append(obs_pixel)

        obs = torch.cat([ob for ob in obs_list if ob is not None], dim=-1)
            
        if self.rnn_config.if_rnn_policy:
            obs, _ = self.rnn_stem(obs, cells)
            obs = obs.contiguous()

        action = self.actor(obs)
        return action

    def forward_critic(self, obs, cells=None):
        '''
            forward pass critic to generate policy with option to use z-filter
            Note: assumes input has either shape length 2 or 4 without RNN and
            length 3 or 5 with RNN depending on if image based training is used
            Args: 
                obs: dictionary of batched observations
                cells: tuple of hidden and cell state for LSTM stem. optional
            Returns:
                output of critic network
        '''
        obs_list = []
        obs_flat = self._gather_low_dim_input(obs)
        if self.use_z_filter:
            obs_flat = self.z_filter.forward(obs_flat)
        obs_list.append(obs_flat)

        if self.if_pixel_input:
            # right now assumes only one camera angle.
            obs_pixel = obs['pixel']['camera0']
            obs_pixel = self._scale_image(obs_pixel)
            obs_pixel = self.cnn_stem(obs_pixel)
            obs_list.append(obs_pixel)

        obs = torch.cat([ob for ob in obs_list if ob is not None], dim=-1)

        if self.rnn_config.if_rnn_policy:
            obs, _ = self.rnn_stem(obs, cells)
            obs = obs.contiguous()

        value = self.critic(obs)
        return value

    def forward_actor_expose_cells(self, obs, cells=None):
        '''
            forward pass critic to generate policy with option to use z-filter
            also returns an updated LSTM hidden/cell state when necessary
            Args: 
                obs: dictionary of batched observations
                cells: tuple of hidden and cell state for LSTM stem. optional
            Returns:
                output of critic network
        '''
        obs_list = []
        obs_flat = self._gather_low_dim_input(obs)
        if self.use_z_filter:
            obs_flat = self.z_filter.forward(obs_flat)
        obs_list.append(obs_flat)

        if self.if_pixel_input:
            # right now assumes only one camera angle.
            obs_pixel = obs['pixel']['camera0']
            obs_pixel = self._scale_image(obs_pixel)
            obs_pixel = self.cnn_stem(obs_pixel) 
            obs_list.append(obs_pixel)

        obs = torch.cat([ob for ob in obs_list if ob is not None], dim=-1)

        if self.rnn_config.if_rnn_policy:
            obs = obs.view(1, 1, -1) # assume input is shape (1, obs_dim)
            obs, cells = self.rnn_stem(obs, cells)
            
            # .detach() is necessary here to prevent overflow of memory
            # otherwise rollout in length of thousands will prevent previously
            # accumulated hidden/cell states from being freed.
            cells = (cells[0].detach(),cells[1].detach())
            obs = obs.contiguous()  
            obs = obs.view(-1, self.rnn_config.rnn_hidden)

        action = self.actor(obs) # shape (1, action_dim)
        return action, cells

    def z_update(self, obs):
        '''
            updates Z_filter running obs mean and variance
            Args:
                obs: dictionary of batched observations
        '''
        if self.use_z_filter:
            obs_flat = self._gather_low_dim_input(obs)
            self.z_filter.z_update(obs_flat)
        else:
            raise ValueError('Z_update called when network is set to not use z_filter')

    def _scale_image(self, obs):
        '''
            Given uint8 input from the environment, scale to float32 and
            divide by 255 to scale inputs between 0.0 and 1.0
            Args:
                obs: dictionary of batched observations
        '''
        return obs / 255.0



import numpy as np
import sys
import collections
import dm_control
from dm_control.suite.wrappers import pixels
from dm_control.rl.environment import StepType
from .wrapper import Wrapper, SpecFormat
from .base import ActionType


class DMControlDummyWrapper(Wrapper):
    
    '''
    Dummy wrapper for deepmind control environment using pixels.  The output of
    observation_spec and action_spec will match the output for a dm_control environment
    using pixels.Wrapper().  This is used by the learner to get the action and observation
    specs without initializing a pixels wrapper.
    '''

    def __init__(self, env):
        # dm_control envs don't have metadata
        env.metadata = {}
        super().__init__(env)

    @property
    def spec_format(self):
        return SpecFormat.DM_CONTROL

    def observation_spec(self):
        modality = collections.OrderedDict([('pixels', dm_control.rl.specs.ArraySpec(shape=(84, 84, 3), dtype=np.dtype('float32')))])
        return modality

    def action_spec(self):
        return self.env.action_spec()

class DMControlAdapter(Wrapper):
    def __init__(self, env, is_pixel_input):
        # dm_control envs don't have metadata
        env.metadata = {}
        super().__init__(env)
        self.screen = None
        self.is_pixel_input = is_pixel_input
        assert (isinstance(env, dm_control.rl.control.Environment) or
            isinstance(env, pixels.Wrapper) or
            isinstance(env, DMControlDummyWrapper))

    def _add_modality(self, obs):
        if self.is_pixel_input:
            renamed_obs = collections.OrderedDict([('camera0', obs['pixels'])])
            return collections.OrderedDict([('pixel', renamed_obs)])
        else:
            return collections.OrderedDict([('low_dim', obs)])

    def _step(self, action):
        ts = self.env.step(action)
        reward = ts.reward
        if reward is None:
            # TODO: note that reward is none
            print('None reward')
            reward = 0
        return self._add_modality(ts.observation), reward, ts.step_type == StepType.LAST, {}

    def _reset(self):
        ts = self.env.reset()
        return self._add_modality(ts.observation), {}

    def _close(self):
        self.env.close()

    @property
    def spec_format(self):
        return SpecFormat.SpecFormat.DM_CONTROL

    def observation_spec(self):
        obs_spec = collections.OrderedDict()
        for modality, v in self._add_modality(self.env.observation_spec()).items():
            modality_spec = collections.OrderedDict()
            for key, obs_shape in v.items():
                # Deepmind observation spec uses Deepmind data types, we just need shape as tuple
                modality_spec[key] = obs_shape.shape
            obs_spec[modality] = modality_spec
        return obs_spec

    def action_spec(self):
        return {
            'type': ActionType.continuous,
            'dim': self.env.action_spec().shape, # DM_control returns int, we want all dim to be tuple
        }

    def _render(self, *args, width=480, height=480, camera_id=1, **kwargs):
        # safe for multiple calls
        import pygame
        pygame.init()
        if not self.screen:
            self.screen = pygame.display.set_mode((width, height))
        else:
            c_width, c_height = self.screen.get_size()
            if c_width != width or c_height != height:
                self.screen = pygame.display.set_mode((width, height))
        for event in pygame.event.get():
            if event.type == pygame.QUIT: sys.exit()

        im = self.env.physics.render(width=width,
            height=height, camera_id=camera_id).transpose((1,0,2))
        pygame.pixelcopy.array_to_surface(self.screen, im)
        pygame.display.update()
        return im


import argparse
from surreal.session import (
    Config,
    LOCAL_SESSION_CONFIG,
    BASE_LEARNER_CONFIG,
    BASE_ENV_CONFIG
    )
from surreal.agent import DDPGAgent
from surreal.learner import DDPGLearner
from surreal.replay import UniformReplay
from surreal.launch import SurrealDefaultLauncher
from surreal.env import make_env

# TODO：Documentation on config files

DDPG_DEFAULT_LEARNER_CONFIG = Config({
    'model': {
        'convs': [],
        'actor_fc_hidden_sizes': [300, 200],
        'critic_fc_hidden_sizes': [400, 300],
        'use_layernorm': False,
        'conv_spec': {
            # First conv layer: 16 out channels, second layer 32 channels
            'out_channels': [16, 32],
            # First conv layer: kernel size 8, second layer kernel size 4
            'kernel_sizes': [8, 4],
            # First conv layer: stride=4, second layer stride=2
            'strides': [4, 2],
            # After final convolution, reshapes output to flat tensor and feed through mlp with output of this size
            'hidden_output_dim': 200,
        },
    },
    'algo': {
        'gamma': .99,
        # Unroll the bellman update
        'n_step': 3,
        # Send experiences every `stride` steps
        'stride': 1,
        'network': {
            'lr_actor': 1e-4,
            'lr_critic': 1e-3,
            'clip_actor_gradient': True,
            'actor_gradient_value_clip': 1.,
            'clip_critic_gradient': False,
            'critic_gradient_value_clip': 5.,
            # Weight regularization
            'actor_regularization': 0.0,
            'critic_regularization': 0.0,
            # beta version: see https://arxiv.org/pdf/1802.09477.pdf and
            # https://github.com/sfujim/TD3/blob/master/TD3.py
            # for action regularization and double critic algorithm details
            'use_action_regularization': False,
            'use_double_critic': False,
            'target_update': {
                # Soft: after every iteration, target_params = (1 - tau) * target_params + tau * params
                #'type': 'soft',
                #'tau': 1e-3,
                # Hard: after `interval` iterations, target_params = params
                'type': 'hard',
                'interval': 500,
            },
        },
        'exploration': {
            # Beta implementation of parameter noise:
            # see https://blog.openai.com/better-exploration-with-parameter-noise/ for algorithm details
            'param_noise_type': None,

            # normal parameter noise applies gaussian noise over the agent's parameters
            # 'param_noise_type': 'normal',

            # adaptive parameter noise scales the noise sigma up or down in order to achieve the target action
            # standard deviation
            # 'param_noise_type': 'adaptive_normal',
            'param_noise_sigma': 0.05,
            'param_noise_alpha': 1.15,
            'param_noise_target_stddev': 0.005,

            # Vanilla noise: applies gaussian noise on every action
            'noise_type': 'normal',
            'max_sigma': 1.0,

            # Or, use Ornstein-Uhlenbeck noise instead of gaussian
            #'noise_type': 'ou_noise',
            'theta': 0.15,
            'dt': 1e-3,
        },
    },
    'replay': {
        'batch_size': 512,
        'memory_size': int(1000000/3), # The total replay size is memory_size * replay_shards
        'sampling_start_size': 3000,
        'replay_shards': 3,
    },
    'parameter_publish': {
        # Minimum amount of time (seconds) between two parameter publish
        'min_publish_interval': 3,
    },
})

DDPG_DEFAULT_LEARNER_CONFIG.extend(BASE_LEARNER_CONFIG)

DDPG_DEFAULT_ENV_CONFIG = Config({
    'env_name': '_str_',
    'num_agents': '_int_',

    'demonstration': None,
    'use_depth': False,

    'use_demonstration': False,
    # If true, DDPG will expect an image at obs['pixel']['camera0']
    'pixel_input': False,
    'use_grayscale': False,
    # Stacks previous image frames together to provide history information
    'frame_stacks': 3,
    # Each action will be played this number of times. The reward of the consecutive actions will be the the reward
    # of the last action in the sequence
    'action_repeat': 1,
    # If false, the agent will send an image will be a list of frames to the replay.  When the learner receives an
    # observation, it will concatenate the frames into a single tensor.  This allows the replay to optimize memory
    # usage so that identical frames aren't duplicated in memory
    'frame_stack_concatenate_on_env': False,
    # Debug only: agent will sleep for this number of seconds between actions
    'sleep_time': 0.0,
    # If an episode reaches this number of steps, the state will be considered terminal
    'limit_episode_length': 0, # 0 means no limit
    'video': {
        'record_video': True,
        'save_folder': None,
        'max_videos': 500,
        'record_every': 20,
    },
    # observation: if using FilterWrapper, any input not listed will be thrown out.
    # For example, if an observation had a value at obs['pixel']['badkey'], that value will be ignored
    'observation': {
        'pixel':['camera0', 'depth'],
        # if using ObservationConcatWrapper, all low_dim inputs will be concatenated together into a single input
        # named 'flat_inputs'
        'low_dim':['position', 'velocity', 'proprio', 'robot-state', 'cube_pos', 'cube_quat', 'gripper_to_cube', 'low-dim'],
    },
})

DDPG_DEFAULT_ENV_CONFIG.extend(BASE_ENV_CONFIG)

DDPG_DEFAULT_SESSION_CONFIG = Config({
    'folder': '_str_',
    'tensorplex': {
        'update_schedule': {
            # for TensorplexWrapper:
            'training_env': 20,  # env record every N episodes
            'eval_env': 5,
            'eval_env_sleep': 30,  # throttle eval by sleep n seconds
            # for manual updates:
            'agent': 50,  # agent.tensorplex.add_scalars()
            'learner': 20,  # learner.tensorplex.add_scalars()
        }
    },
    'agent': {
        # fetch_parameter_mode: 'episode', 'episode:<n>', 'step', 'step:<n>'
        # every episode, every n episodes, every step, every n steps
        'fetch_parameter_mode': 'step',
        'fetch_parameter_interval': 200,
        'num_gpus': 0,
    },
    'sender': {
        'flush_iteration': 100,
    },
    'learner': {
        'prefetch_processes': 3,
        'num_gpus': 0,
    },
})

DDPG_DEFAULT_SESSION_CONFIG.extend(LOCAL_SESSION_CONFIG)

DDPG_BLOCK_LIFTING_LEARNER_CONFIG = Config({
    'algo': {
        'network': {
            'lr_actor': 1e-4,
            'lr_critic': 1e-4,
            # Weight regularization
            'actor_regularization': 1e-4,
            'critic_regularization': 1e-4,
            'target_update': {
                # Soft: after every iteration, target_params = (1 - tau) * target_params + tau * params
                #'type': 'soft',
                #'tau': 1e-3,
                # Hard: after `interval` iterations, target_params = params
                'type': 'hard',
                'interval': 500,
            },
        },
        'exploration': {
            'max_sigma': 2.0,
            # Use Ornstein-Uhlenbeck noise instead of gaussian
            'noise_type': 'ou_noise',
            'theta': 0.15,
            'dt': 1e-3,
        },
    },
})

DDPG_BLOCK_LIFTING_LEARNER_CONFIG.extend(DDPG_DEFAULT_LEARNER_CONFIG)

DDPG_BLOCK_LIFTING_ENV_CONFIG = Config({
    'env_name': '_str_',
    'num_agents': '_int_',
    # If true, DDPG will expect an image at obs['pixel']['camera0']
    'pixel_input': True,
    'use_grayscale': False,
    # Stacks previous image frames together to provide history information
    'frame_stacks': 3,
    # Each action will be played this number of times. The reward of the consecutive actions will be the the reward
    # of the last action in the sequence
    'action_repeat': 10,
    # If an episode reaches this number of steps, the state will be considered terminal
    'limit_episode_length': 200, # 0 means no limit
    # observation: if using FilterWrapper, any input not listed will be thrown out.
    # For example, if an observation had a value at obs['pixel']['badkey'], that value will be ignored
    'observation': {
        'pixel':['camera0', 'depth'],
        # if using ObservationConcatWrapper, all low_dim inputs will be concatenated together into a single input
        # named 'flat_inputs'
        'low_dim':['position', 'velocity', 'robot-state', 'proprio', 'cube_pos', 'cube_quat', 'gripper_to_cube', 'low-dim'],
    },
})

DDPG_BLOCK_LIFTING_ENV_CONFIG.extend(DDPG_DEFAULT_ENV_CONFIG)

class DDPGLauncher(SurrealDefaultLauncher):
    def __init__(self):
        learner_class = DDPGLearner
        agent_class = DDPGAgent
        replay_class = UniformReplay
        learner_config = DDPG_DEFAULT_LEARNER_CONFIG
        env_config = DDPG_DEFAULT_ENV_CONFIG
        session_config = DDPG_DEFAULT_SESSION_CONFIG
        super().__init__(agent_class,
                         learner_class,
                         replay_class,
                         session_config,
                         env_config,
                         learner_config)

    def setup(self, argv):
        """
        The function name must be `generate`.
        Will be called by `surreal.main_scripts.runner`
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--env', type=str, required=True, help='name of the environment')
        parser.add_argument('--num-agents', type=int, required=True, help='number of agents used')
        parser.add_argument('--num-gpus', type=int, default=0,
                            help='number of GPUs to use, 0 for CPU only.')
        parser.add_argument('--agent-num-gpus', type=int, default=0,
                            help='number of GPUs to use for agent, 0 for CPU only.')
        parser.add_argument('--restore-folder', type=str, default=None,
                            help='folder containing checkpoint to restore from')
        parser.add_argument('--experiment-folder', required=True,
                            help='session_config.folder that has experiment files'
                            ' like checkpoint and logs')
        parser.add_argument('--agent-batch', type=int, default=1,
                            help='how many agents/evals per batch')
        parser.add_argument('--eval-batch', type=int, default=1,
                            help='how many agents/evals per batch')
        parser.add_argument('--unit-test', action='store_true',
                            help='Prevents sharding replay and paramter '
                            'server. Helps prevent address collision'
                            ' in unit testing.')

        args = parser.parse_args(args=argv)

        if args.env == 'robosuite:SawyerLift':
            learner_class = DDPGLearner
            agent_class = DDPGAgent
            replay_class = UniformReplay
            learner_config = DDPG_BLOCK_LIFTING_LEARNER_CONFIG
            env_config = DDPG_BLOCK_LIFTING_ENV_CONFIG
            session_config = DDPG_DEFAULT_SESSION_CONFIG
            super.__init__(agent_class,
                           learner_class,
                           replay_class,
                           session_config,
                           env_config,
                           learner_config)

        self.env_config.env_name = args.env
        _, self.env_config = make_env(self.env_config)
        self.env_config.num_agents = args.num_agents

        self.session_config.folder = args.experiment_folder
        self.session_config.agent.num_gpus = args.agent_num_gpus
        self.session_config.learner.num_gpus = args.num_gpus
        if args.restore_folder is not None:
            self.session_config.checkpoint.restore = True
            self.session_config.checkpoint.restore_folder = args.restore_folder
        self.agent_batch_size = args.agent_batch
        self.eval_batch_size = args.eval_batch

        # Used in tests: Prevent IP address in use error
        #                Prevent replay from hanging learner
        #                due to sample_start
        if args.unit_test:
            self.learner_config.replay.sampling_start_size = 5
            self.learner_config.replay.replay_shards = 1
            self.session_config.ps.shards = 1


def main():
    DDPGLauncher().main()

if __name__ == '__main__':
    main()




import shutil
import pkg_resources
import surreal.utils as U


def main():
    print('Generating default config for surreal at ~/.surreal.yml')
    default_path = U.f_expand('~/.surreal.yml')
    U.move_with_backup(default_path)

    fname = pkg_resources.resource_filename('surreal', 'sample_surreal.yml')
    shutil.copyfile(fname, default_path)


if __name__ == "__main__":
    main()



import time
import os
import surreal.utils as U
from surreal.session import get_tensorplex_client, get_loggerplex_client
from surreal.distributed import ExperienceCollectorServer
from caraml.zmq import ZmqServer


class Replay:
    """
        Important: When extending this class, make sure to follow the init
        method signature so that orchestrating functions can properly
        initialize the replay server.
    """
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 index=0):
        """
        """
        # Note that there're 2 replay configs:
        # one in learner_config that controls algorithmic
        # part of the replay logic
        # one in session_config that controls system settings
        self.learner_config = learner_config
        self.env_config = env_config
        self.session_config = session_config
        self.index = index

        collector_port = os.environ['SYMPH_COLLECTOR_BACKEND_PORT']
        sampler_port = os.environ['SYMPH_SAMPLER_BACKEND_PORT']
        self._collector_server = ExperienceCollectorServer(
            host='localhost',
            port=collector_port,
            exp_handler=self._insert_wrapper,
            load_balanced=True,
        )
        self._sampler_server = ZmqServer(
            host='localhost',
            port=sampler_port,
            bind=False)
        self._sampler_server_thread = None

        self._evict_interval = self.session_config.replay.evict_interval
        self._evict_thread = None

        self._setup_logging()

    def start_threads(self):
        if self._has_tensorplex:
            self.start_tensorplex_thread()

        self._collector_server.start()

        if self._evict_interval:
            self.start_evict_thread()

        self._sampler_server_thread = self._sampler_server.start_loop(
            handler=self._sample_request_handler)

    def join(self):
        self._collector_server.join()
        self._sampler_server_thread.join()
        if self._has_tensorplex:
            self._tensorplex_thread.join()
        if self._evict_interval:
            self._evict_thread.join()

    def insert(self, exp_dict):
        """
        Add a new experience to the replay.
        Includes passive evict logic if memory capacity is exceeded.

        Args:
            exp_dict: {[obs], action, reward, done, info}
        """
        raise NotImplementedError

    def sample(self, batch_size):
        """
        This function is called in _sample_handler for learner side Zmq request

        Args:
            batch_size:

        Returns:
            a list of exp_tuples
        """
        raise NotImplementedError

    def evict(self):
        """
        Actively evict old experiences.
        """
        pass

    def start_sample_condition(self):
        """
        Tells the thread to start sampling only when this condition is met.
        For example, only when the replay memory has > 10K experiences.

        Returns:
            bool: whether to start sampling or not
        """
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    # ======================== internal methods ========================
    def _setup_logging(self):
        self.log = get_loggerplex_client(
            '{}/{}'.format('replay', self.index),
            self.session_config
        )
        self.tensorplex = get_tensorplex_client(
            '{}/{}'.format('replay', self.index),
            self.session_config
        )
        self._tensorplex_thread = None
        self._has_tensorplex = self.session_config.replay.tensorboard_display

        # Origin of all global steps
        self.init_time = time.time()
        # Number of experience collected by agents
        self.cumulative_collected_count = 0
        # Number of experience sampled by learner
        self.cumulative_sampled_count = 0
        # Number of sampling requests from the learner
        self.cumulative_request_count = 0
        # Timer for tensorplex reporting
        self.last_tensorplex_iter_time = time.time()
        # Last reported values used for speed computation
        self.last_experience_count = 0
        self.last_sample_count = 0
        self.last_request_count = 0

        self.insert_time = U.TimeRecorder(decay=0.99998)
        self.sample_time = U.TimeRecorder()
        self.serialize_time = U.TimeRecorder()

        # moving avrage of about 100s
        self.exp_in_speed = U.MovingAverageRecorder(decay=0.99)
        self.exp_out_speed = U.MovingAverageRecorder(decay=0.99)
        self.handle_sample_request_speed = U.MovingAverageRecorder(decay=0.99)

    def _insert_wrapper(self, exp):
        """
            Allows us to do some book keeping in the base class
        """
        self.cumulative_collected_count += 1
        with self.insert_time.time():
            self.insert(exp)

    def _sample_request_handler(self, req):
        """
        Handle requests to the learner
        https://stackoverflow.com/questions/29082268/python-time-sleep-vs-event-wait
        Since we don't have external notify, we'd better just use sleep
        """
        batch_size = U.deserialize(req)
        U.assert_type(batch_size, int)
        while not self.start_sample_condition():
            time.sleep(0.01)
        self.cumulative_sampled_count += batch_size
        self.cumulative_request_count += 1
        with self.sample_time.time():
            sample = self.sample(batch_size)
        with self.serialize_time.time():
            return U.serialize(sample)

    def start_evict_thread(self):
        if self._evict_thread is not None:
            raise RuntimeError('evict thread already running')
        self._evict_thread = U.start_thread(self._evict_loop)
        return self._evict_thread

    def _evict_loop(self):
        assert self._evict_interval
        while True:
            time.sleep(self._evict_interval)
            self.evict()

    def start_tensorplex_thread(self):
        if self._tensorplex_thread is not None:
            raise RuntimeError('tensorplex thread already running')
        self._tensorplex_thread = U.PeriodicWakeUpWorker(
            target=self.generate_tensorplex_report)
        self._tensorplex_thread.start()
        return self._tensorplex_thread

    def generate_tensorplex_report(self):
        """
            Generates stats to be reported to tensorplex
        """
        global_step = int(time.time() - self.init_time)

        time_elapsed = time.time() - self.last_tensorplex_iter_time + 1e-6

        cum_count_collected = self.cumulative_collected_count
        new_exp_count = cum_count_collected - self.last_experience_count
        self.last_experience_count = cum_count_collected

        cum_count_sampled = self.cumulative_sampled_count
        new_sample_count = cum_count_sampled - self.last_sample_count
        self.last_sample_count = cum_count_sampled

        cum_count_requests = self.cumulative_request_count
        new_request_count = cum_count_requests - self.last_request_count
        self.last_request_count = cum_count_requests

        exp_in_speed = self.exp_in_speed.add_value(new_exp_count / time_elapsed)
        exp_out_speed = self.exp_out_speed.add_value(new_sample_count / time_elapsed)
        handle_sample_request_speed = self.handle_sample_request_speed.add_value(
                                                    new_request_count / time_elapsed)

        insert_time = self.insert_time.avg
        sample_time = self.sample_time.avg
        serialize_time = self.serialize_time.avg

        core_metrics = {
            'num_exps': len(self),
            'total_collected_exps': cum_count_collected,
            'total_sampled_exps': cum_count_sampled,
            'total_sample_requests': self.cumulative_request_count,
            'exp_in_per_s': exp_in_speed,
            'exp_out_per_s': exp_out_speed,
            'requests_per_s': handle_sample_request_speed,
            'insert_time_s': insert_time,
            'sample_time_s': sample_time,
            'serialize_time_s': serialize_time,
        }

        serialize_load = serialize_time * handle_sample_request_speed / time_elapsed
        collect_exp_load = insert_time * exp_in_speed / time_elapsed
        sample_exp_load = sample_time * handle_sample_request_speed / time_elapsed

        system_metrics = {
            'lifetime_experience_utilization_percent':
            cum_count_sampled / (cum_count_collected + 1) * 100,
            'current_experience_utilization_percent': exp_out_speed / (exp_in_speed + 1) * 100,
            'serialization_load_percent': serialize_load * 100,
            'collect_exp_load_percent': collect_exp_load * 100,
            'sample_exp_load_percent': sample_exp_load * 100,
            # 'exp_queue_occupancy_percent': self._exp_queue.occupancy() * 100,
        }

        all_metrics = {}
        for k in core_metrics:
            all_metrics['.core/' + k] = core_metrics[k]
        for k in system_metrics:
            all_metrics['.system/' + k] = system_metrics[k]
        self.tensorplex.add_scalars(all_metrics, global_step=global_step)

        self.last_tensorplex_iter_time = time.time()
